#!/usr/bin/env python
# -*- coding: ISO-8859-1 -*-
#
#
# Copyright (C) 2002-2004 Jörg Lehmann <joergl@users.sourceforge.net>
# Copyright (C) 2003-2004 Michael Schindler <m-schindler@users.sourceforge.net>
# Copyright (C) 2002-2004 André Wobst <wobsta@users.sourceforge.net>
#
# This file is part of PyX (http://pyx.sourceforge.net/).
#
# PyX is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PyX is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyX; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


import re, ConfigParser
from pyx import mathtree
from pyx.graph import style


class _Idata:
    """interface definition of a data object
    data objects store data arranged in rows and columns"""

    columns = {}
    """a dictionary mapping column titles to column numbers"""

    points = []
    """column/row data
    - a list of rows where each row represents a data point
    - each row contains a list, where each entry of the list represents a value for a column
    - the number of columns for each data point must match the number of columns
    - any column enty of any data point might be a float, a string, or None"""

    title = ""
    """a string (for printing in PyX, e.g. in a graph key)
    - None is allowed, which marks the data instance to have no title,
      e.g. it should be skiped in a graph key etc.
    - the title does need to be unique"""

    def getcolumnnumber(self, column):
        """returns a column number
        - the column parameter might be an integer to be used as a column number
        - a column number must be a valid list index (negative values are allowed)
        - the column parameter might be a string contained in the columns list;
          to be valid, the string must be unique within the columns list"""

    def getcolumn(self, column):
        """returns a column
        - extracts a column out of self.data and returns it as a list
        - the column is identified by the parameter column as in getcolumnnumber"""


class _data:

    defaultstyle = style.symbol()

    def getcolumnnumber(self, key):
        try:
            key + ""
        except:
            return key + 0
        else:
            return self.columns[key.strip()]

    def getcolumn(self, key):
        columnno = self.getcolumnnumber(key)
        return [point[columnno] for point in self.points]

    def setstyle(self, graph, style):
        self.style = style
        unhandledcolumns = self.style.setdata(graph, self.columns, self)
        unhandledcolumnkeys = unhandledcolumns.keys()
        if len(unhandledcolumnkeys):
            raise ValueError("style couldn't handle column keys %s" % unhandledcolumnkeys)

    def selectstyle(self, graph, selectindex, selecttotal):
        self.style.selectstyle(selectindex, selecttotal, self)

    def adjustaxes(self, graph, step):
        """
        - on step == 0 axes with fixed data should be adjusted
        - on step == 1 the current axes ranges might be used to
          calculate further data (e.g. y data for a function y=f(x)
          where the y range depends on the x range)
        - on step == 2 axes ranges not previously set should be
          updated by data accumulated by step 1"""
        if step == 0:
            self.style.adjustaxes(self.columns.values(), self)

    def draw(self, graph):
        self.style.drawpoints(graph, self)


class list(_data):
    "creates data out of a list"

    def checkmaxcolumns(self, points, maxcolumns=None):
        if maxcolumns is None:
            maxcolumns = max([len(point) for point in points])
        for i in xrange(len(points)):
            l = len(points[i])
            if l < maxcolumns:
                try:
                    p = points[i] + [None] * (maxcolumns - l)
                except:
                    # points[i] are not a list
                    p = __builtins__.list(points[i]) + [None] * (maxcolumns - l)
                try:
                    points[i] = p
                except:
                    # points are not a list -> end loop without step into else
                    break
        else:
            # the loop finished successfull
            return points
        # since points are not a list, convert them and try again
        return checkmaxcolumns(__builtins__.list(points), maxcolumns=maxcolumns)

    def __init__(self, points, title="user provided list", maxcolumns=None, addlinenumbers=1, **columns):
        points = self.checkmaxcolumns(points, maxcolumns)
        if addlinenumbers:
            for i in xrange(len(points)):
                try:
                    points[i].insert(0, i+1)
                except:
                    points[i] = [i+1] + __builtins__.list(points[i])
        self.points = points
        self.columns = columns
        self.title = title


##############################################################
# math tree enhanced by column handling
##############################################################

class MathTreeFuncCol(mathtree.MathTreeFunc1):

    def __init__(self, *args):
        mathtree.MathTreeFunc1.__init__(self, "_column_", *args)

    def VarList(self):
        # we misuse VarList here:
        # - instead of returning a string, we return this instance itself
        # - before calculating the expression, you must call ColumnNameAndNumber
        #   once (when limiting the context to external defined variables,
        #   otherwise you have to call it each time)
        return [self]

    def ColumnNameAndNumber(_hidden_self, **args):
        number = int(_hidden_self.Args[0].Calc(**args))
        _hidden_self.varname = "_column_%i" % number
        return _hidden_self.varname, number

    def __str__(self):
        return self.varname

    def Calc(_hidden_self, **args):
        return args[_hidden_self.varname]

MathTreeFuncsWithCol = mathtree.DefaultMathTreeFuncs + [MathTreeFuncCol]


class columntree:

    def __init__(self, tree):
        self.tree = tree
        self.Calc = tree.Calc
        self.__str__ = tree.__str__

    def VarList(self):
        # returns a list of regular variables (strings) like the original mathtree
        return [var for var in self.tree.VarList() if not isinstance(var, MathTreeFuncCol) and var[:8] != "_column_"]

    def columndict(_hidden_self, **context):
        # returns a dictionary of column names (keys) and column numbers (values)
        columndict = {}
        for var in _hidden_self.tree.VarList():
            if isinstance(var, MathTreeFuncCol):
                name, number = var.ColumnNameAndNumber(**context)
                columndict[name] = number
            elif var[:8] == "_column_":
                columndict[var] = int(var[8:])
        return columndict


class dataparser(mathtree.parser):
    # mathtree parser enhanced by column handling
    # parse returns a columntree instead of a regular tree

    def __init__(self, MathTreeFuncs=MathTreeFuncsWithCol, **kwargs):
        mathtree.parser.__init__(self, MathTreeFuncs=MathTreeFuncs, **kwargs)

    def parse(self, expr):
        return columntree(mathtree.parser.parse(self, expr.replace("$", "_column_")))

##############################################################


class copycolumn:
    # a helper storage class to mark a new column to copied
    # out of data from an old column
    def __init__(self, newcolumntitle, oldcolumnnumber):
        self.newcolumntitle = newcolumntitle
        self.oldcolumnnumber = oldcolumnnumber

class mathcolumn:
    """a helper storage class to mark a new column to created
    by evaluating a mathematical expression"""
    def __init__(self, newcolumntitle, expression, tree, varitems):
        # - expression is a string
        # - tree is a parsed mathematical tree, e.g. we can have
        #   call tree.Calc(**vars), where the dict vars maps variable
        #   names to values
        # - varitems is a list of (key, value) pairs, where the key
        #   stands is a variable name in the mathematical tree and
        #   the value is its value"""
        self.newcolumntitle = newcolumntitle
        self.expression = expression
        self.tree = tree
        self.varitems = varitems

class notitle:
    """this is a helper class to mark, that no title was privided
    (since a title equals None is a valid input, it needs to be
    distinguished from providing no title when a title will be
    created automatically)"""
    pass

class data(_data):
    "creates a new data set out of an existing data set"

    def __init__(self, data, title=notitle, parser=dataparser(), context={}, **columns):
        defaultstyle = data.defaultstyle

        # build a nice title
        if title is notitle:
            items = columns.items()
            items.sort() # we want sorted items (otherwise they would be unpredictable scrambled)
            self.title = data.title + ": " + ", ".join(["%s=%s" % item for item in items])
        else:
            self.title = title

        # analyse the **columns argument
        newcolumns = []
        hasmathcolumns = 0
        for newcolumntitle, columnexpr in columns.items():
            try:
                # try if it is a valid column identifier
                oldcolumnnumber = data.getcolumnnumber(columnexpr)
            except:
                # if not it should be a mathematical expression
                tree = parser.parse(columnexpr)
                columndict = tree.columndict(**context)
                for var in tree.VarList():
                    try:
                        columndict[var] = data.getcolumnnumber(var)
                    except KeyError, e:
                        if var not in context.keys():
                            raise e
                newcolumns.append(mathcolumn(newcolumntitle, columnexpr, tree, columndict.items()))
                hasmathcolumns = 1
            else:
                newcolumns.append(copycolumn(newcolumntitle, oldcolumnnumber))

        # ensure to copy the zeroth column (line number)
        # if we already do, place it first again, otherwise add it to the front
        i = 0
        for newcolumn in newcolumns:
            if isinstance(newcolumn, copycolumn) and not newcolumn.oldcolumnnumber:
                newcolumns.pop(i)
                newcolumns.insert(0, newcolumn)
                firstcolumnwithtitle = 0
                break
            i += 1
        else:
            newcolumns.insert(0, copycolumn(None, 0))
            firstcolumnwithtitle = 1

        if hasmathcolumns:
            # new column data needs to be calculated
            vars = context.copy() # do not modify context, use a copy vars instead
            self.points = [None]*len(data.points)
            countcolumns = len(newcolumns)
            for i in xrange(len(data.points)):
                datapoint = data.points[i]
                point = [None]*countcolumns
                newcolumnnumber = 0
                for newcolumn in newcolumns:
                    if isinstance(newcolumn, copycolumn):
                        point[newcolumnnumber] = datapoint[newcolumn.oldcolumnnumber]
                    else:
                        # update the vars
                        # TODO: we could update it once for all varitems
                        for newcolumntitle, value in newcolumn.varitems:
                            vars[newcolumntitle] = datapoint[value]
                        point[newcolumnnumber] = newcolumn.tree.Calc(**vars)
                        # we could also do:
                        # point[newcolumnnumber] = eval(str(newcolumn.tree), vars)
                    newcolumnnumber += 1
                self.points[i] = point

            # store the column titles
            self.columns = {}
            newcolumnnumber = firstcolumnwithtitle
            for newcolumn in newcolumns[firstcolumnwithtitle:]:
                self.columns[newcolumn.newcolumntitle] = newcolumnnumber
                newcolumnnumber += 1
        else:
            # since only column copies are needed, we can share the original points
            self.points = data.points

            # store the new column titles
            self.columns = {}
            for newcolumn in newcolumns[firstcolumnwithtitle:]:
                self.columns[newcolumn.newcolumntitle] = newcolumn.oldcolumnnumber


filecache = {}

class file(data):

    defaultcommentpattern = re.compile(r"(#+|!+|%+)\s*")
    defaultstringpattern = re.compile(r"\"(.*?)\"(\s+|$)")
    defaultcolumnpattern = re.compile(r"(.*?)(\s+|$)")

    def splitline(self, line, stringpattern, columnpattern, tofloat=1):
        """returns a tuple created out of the string line
        - matches stringpattern and columnpattern, adds the first group of that
          match to the result and and removes those matches until the line is empty
        - when stringpattern matched, the result is always kept as a string
        - when columnpattern matched and tofloat is true, a conversion to a float
          is tried; when this conversion fails, the string is kept"""
        result = []
        # try to gain speed by skip matching regular expressions
        if line.find('"')!=-1 or \
           stringpattern is not self.defaultstringpattern or \
           columnpattern is not self.defaultcolumnpattern:
            while len(line):
                match = stringpattern.match(line)
                if match:
                    result.append(match.groups()[0])
                    line = line[match.end():]
                else:
                    match = columnpattern.match(line)
                    if tofloat:
                        try:
                            result.append(float(match.groups()[0]))
                        except (TypeError, ValueError):
                            result.append(match.groups()[0])
                    else:
                        result.append(match.groups()[0])
                    line = line[match.end():]
        else:
            if tofloat:
                try:
                    return map(float, line.split())
                except (TypeError, ValueError):
                    result = []
                    for r in line.split():
                        try:
                            result.append(float(r))
                        except (TypeError, ValueError):
                            result.append(r)
            else:
                return line.split()
        return result

    def getcachekey(self, *args):
        return ":".join([str(x) for x in args])

    def __init__(self, filename,
                       commentpattern=defaultcommentpattern,
                       stringpattern=defaultstringpattern,
                       columnpattern=defaultcolumnpattern,
                       skiphead=0, skiptail=0, every=1,
                       **kwargs):
        cachekey = self.getcachekey(filename, commentpattern, stringpattern, columnpattern, skiphead, skiptail, every)
        if not filecache.has_key(cachekey):
            file = open(filename)
            self.title = filename
            columns = {}
            points = []
            linenumber = 0
            maxcolumns = 0
            for line in file.readlines():
                line = line.strip()
                match = commentpattern.match(line)
                if match:
                    if not len(points):
                        keys = self.splitline(line[match.end():], stringpattern, columnpattern, tofloat=0)
                        i = 0
                        for key in keys:
                            i += 1
                            columns[key] = i
                else:
                    linedata = []
                    for value in self.splitline(line, stringpattern, columnpattern, tofloat=1):
                        linedata.append(value)
                    if len(linedata):
                        if linenumber >= skiphead and not ((linenumber - skiphead) % every):
                            linedata = [linenumber + 1] + linedata
                            if len(linedata) > maxcolumns:
                                maxcolumns = len(linedata)
                            points.append(linedata)
                        linenumber += 1
            if skiptail:
                del points[-skiptail:]
            filecache[cachekey] = list(points, title=filename, maxcolumns=maxcolumns, addlinenumbers=0, **columns)
        data.__init__(self, filecache[cachekey], **kwargs)


conffilecache = {}

class conffile(data):

    def __init__(self, filename, **kwargs):
        """read data from a config-like file
        - filename is a string
        - each row is defined by a section in the config-like file (see
          config module description)
        - the columns for each row are defined by lines in the section file;
          the option entries identify and name the columns
        - further keyword arguments are passed to the constructor of data,
          keyword arguments data and titles excluded"""
        cachekey = filename
        if not filecache.has_key(cachekey):
            config = ConfigParser.ConfigParser()
            config.optionxform = str
            config.readfp(open(filename, "r"))
            sections = config.sections()
            sections.sort()
            points = [None]*len(sections)
            maxcolumns = 1
            columns = {}
            for i in xrange(len(sections)):
                point = [sections[i]] + [None]*(maxcolumns-1)
                for option in config.options(sections[i]):
                    value = config.get(sections[i], option)
                    try:
                        value = float(value)
                    except:
                        pass
                    try:
                        index = columns[option]
                    except KeyError:
                        columns[option] = maxcolumns
                        point.append(value)
                        maxcolumns += 1
                    else:
                        point[index] = value
                points[i] = point
            conffilecache[cachekey] = list(points, title=filename, maxcolumns=maxcolumns, addlinenumbers=0, **columns)
        data.__init__(self, conffilecache[cachekey], **kwargs)



class function:

    defaultstyle = style.line()

    def __init__(self, expression, title=notitle, min=None, max=None,
    points=100, parser=mathtree.parser(), context={}):

        if title is notitle:
            self.title = expression
        else:
            self.title = title
        self.min = min
        self.max = max
        self.numberofpoints = points
        self.context = context.copy() # be save on late evaluations
        self.result, expression = [x.strip() for x in expression.split("=")]
        self.mathtree = parser.parse(expression)
        self.variable = None

    def setstyle(self, graph, style):
        self.style = style
        for variable in self.mathtree.VarList():
            if variable in graph.axes.keys():
                if self.variable is None:
                    self.variable = variable
                else:
                    raise ValueError("multiple variables found")
        if self.variable is None:
            raise ValueError("no variable found")
        self.xaxis = graph.axes[self.variable]
        self.columns = {self.variable: 1, self.result: 2}
        unhandledcolumns = self.style.setdata(graph, self.columns, self)
        unhandledcolumnkeys = unhandledcolumns.keys()
        if len(unhandledcolumnkeys):
            raise ValueError("style couldn't handle column keys %s" % unhandledcolumnkeys)

    def selectstyle(self, graph, selectindex, selecttotal):
        self.style.selectstyle(selectindex, selecttotal, self)

    def adjustaxes(self, graph, step):
        """
        - on step == 0 axes with fixed data should be adjusted
        - on step == 1 the current axes ranges might be used to
          calculate further data (e.g. y data for a function y=f(x)
          where the y range depends on the x range)
        - on step == 2 axes ranges not previously set should be
          updated by data accumulated by step 1"""
        if step == 0:
            self.points = []
            if self.min is not None:
                self.points.append([None, self.min])
            if self.max is not None:
                self.points.append([None, self.max])
            self.style.adjustaxes([1], self)
        elif step == 1:
            min, max = graph.axes[self.variable].getrange()
            if self.min is not None: min = self.min
            if self.max is not None: max = self.max
            vmin = self.xaxis.convert(min)
            vmax = self.xaxis.convert(max)
            self.points = []
            for i in range(self.numberofpoints):
                v = vmin + (vmax-vmin)*i / (self.numberofpoints-1.0)
                x = self.xaxis.invert(v)
                # caution: the virtual coordinate might differ once
                # the axis rescales itself to include further ticks etc.
                self.points.append([v, x, None])
            for point in self.points:
                self.context[self.variable] = point[1]
                try:
                    point[2] = self.mathtree.Calc(**self.context)
                except (ArithmeticError, ValueError):
                    pass
        elif step == 2:
            self.style.adjustaxes([2], self)

    def draw(self, graph):
        self.style.drawpoints(graph, self)


class paramfunction:

    defaultstyle = style.line()

    def __init__(self, varname, min, max, expression, title=notitle, points=100, parser=mathtree.parser(), context={}):
        if title is notitle:
            self.title = expression
        else:
            self.title = title
        self.varname = varname
        self.min = min
        self.max = max
        self.numberofpoints = points
        self.expression = {}
        varlist, expressionlist = expression.split("=")
        keys = varlist.split(",")
        mathtrees = parser.parse(expressionlist)
        if len(keys) != len(mathtrees):
            raise ValueError("unpack tuple of wrong size")
        self.points = [None]*self.numberofpoints
        emptyresult = [None]*len(keys)
        self.columns = {}
        i = 1
        for key in keys:
            self.columns[key.strip()] = i
            i += 1
        for i in range(self.numberofpoints):
            param = self.min + (self.max-self.min)*i / (self.numberofpoints-1.0)
            context[self.varname] = param
            self.points[i] = [param] + emptyresult
            column = 1
            for key, column in self.columns.items():
                self.points[i][column] = mathtrees[column-1].Calc(**context)
                column += 1

    def setstyle(self, graph, style):
        self.style = style
        unhandledcolumns = self.style.setdata(graph, self.columns, self)
        unhandledcolumnkeys = unhandledcolumns.keys()
        if len(unhandledcolumnkeys):
            raise ValueError("style couldn't handle column keys %s" % unhandledcolumnkeys)

    def selectstyle(self, graph, selectindex, selecttotal):
        self.style.selectstyle(selectindex, selecttotal, self)

    def adjustaxes(self, graph, step):
        if step == 0:
            self.style.adjustaxes(self.columns.values(), self)

    def draw(self, graph):
        self.style.drawpoints(graph, self)




#!/usr/bin/env python
# -*- coding: ISO-8859-1 -*-
#
#
# Copyright (C) 2004 André Wobst <wobsta@users.sourceforge.net>
#
# This file is part of PyX (http://pyx.sourceforge.net/).
#
# PyX is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PyX is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyX; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

_PFB_ASCII = "\200\1"
_PFB_BIN = "\200\2"
_PFB_DONE = "\200\3"
_PFA = "%!"

def fullfont(file, filename):
    """inserts full pfa or pfb fonts
    - file is a file instance where the pfa output is written to
    - pfbfilename is the full filename of the pfa or pfb input file
    - the input type pfa or pfb is autodetected"""

    infile = open(filename, "rb")
    blockid = infile.read(2)
    if blockid == _PFA:
        file.write(blockid)
        file.write(infile.read().replace("\r\n", "\n").replace("\r", "\n"))
    else:
        while 1:
            if len(blockid) != 2:
                raise RuntimeError("EOF reached while reading blockid")
            if blockid == _PFB_DONE:
                if infile.read() != "":
                    raise RuntimeError("tailing characters in pfb file")
                else:
                    return
            if blockid != _PFB_ASCII and blockid != _PFB_BIN:
                raise RuntimeError("invalid blockid")
            length = 0
            for i in range(4):
                length = length + ord(infile.read(1))*(256l**i)
            block = infile.read(length)
            if blockid == _PFB_ASCII:
                file.write(block.replace("\r\n", "\n").replace("\r", "\n"))
            else:
                block = [ord(c) for c in block]
                while len(block) > 32:
                    file.write("%02x"*32 % tuple(block[:32]) + "\n")
                    del block[:32]
                if len(block):
                    file.write("%02x"*len(block) % tuple(block) + "\n")
            blockid = infile.read(2)




# Axis painters performs the painting of an axis. By default, ticks
# are stroked inside of the graph (in the path examples there is no
# graph but don't mind) while labels and the axis title are plotted
# outside. The axis title is rotated along the axis (without writing it
# upside down), while the tick labels are not rotated. The axis painters
# takes a variety of keyword arguments to modify the default
# behaviour.

from pyx import *
from pyx.graph import axis

ap = axis.painter.regular(outerticklength=axis.painter.ticklength.normal)

c = axis.pathaxis(path.curve(0, 0, 3, 0, 1, 4, 4, 4),
                        axis.linear(min=0, max=10, title="axis title",
                                    painter=ap))
c.writeEPSfile("painter")



from pyx import *

c = canvas.canvas()

rect1 = path.path(path.moveto(0, 0), path.lineto(0, 1),
                  path.lineto(1, 1), path.lineto(1, 0),
                  path.lineto(0, 0))

rect2 = path.path(path.moveto(0, 0), path.lineto(0, 1),
                  path.lineto(1, 1), path.lineto(1, 0),
                  path.closepath())

c.stroke(rect1, [trafo.scale(2), style.linewidth.THICK])
c.text(1, -0.7, "not closed", [text.halign.center])
c.stroke(rect2, [trafo.scale(2).translated(4, 0), style.linewidth.THICK])
c.text(5, -0.7, "closed", [text.halign.center])

c.stroke(rect2, [trafo.scale(2).translated(8, 0), style.linewidth.THICK, deco.filled([color.grey(0.95)])])
c.text(9, -0.7, "filled", [text.halign.center])

c.writeEPSfile("rects", paperformat="a4")




import datetime
from pyx.graph import style
from pyx.graph.axis import axis, rater

"""some experimental code for creating a time axis
- it needs python 2.3 to be used (it is based on the new datetime data type)
- a timeaxis is always based on the datetime data type (there is no distinction between times and dates)
"""

class _timemap:
    "time axis mapping based "

    __implements__ = axis._Imap

    def setbasepoints(self, basepoints):
        (self.x1, self.y1), (self.x2, self.y2) = basepoints
        self.dx = self.x2-self.x1
        self.dy = self.y2-self.y1
        if self.dx < datetime.timedelta(0):
            raise RuntimeError("reverse time axis is not expected to work") # and should not happen

    def convert(self, x):
        # XXX float division of timedelta instances
        def mstimedelta(td):
            "return the timedelta in microseconds"
            return td.microseconds + 1000000*(td.seconds + 3600*24*td.days)
        return self.y1 + self.dy * mstimedelta(x - self.x1) / float(mstimedelta(self.dx))
        # we could store float(mstimedelta(self.dx)) instead of self.dx, but
        # I prefer a different solution (not based on huge integers) for the
        # future

    def invert(self, y):
        f = (y - self.y1) / float(self.dy)
        return self.x1 + datetime.timedelta(days=f*self.dx.days, seconds=f*self.dx.seconds, microseconds=f*self.dx.seconds)


class timetick(datetime.datetime):

    def __init__(self, year, month, day, ticklevel=0, labellevel=0, label=None, labelattrs=[], **kwargs):
        datetime.datetime.__init__(self, year, month, day, **kwargs)
        self.ticklevel = ticklevel
        self.labellevel = labellevel
        self.label = label
        self.labelattrs = labelattrs[:]

    def merge(self, other):
        if self.ticklevel is None or (other.ticklevel is not None and other.ticklevel < self.ticklevel):
            self.ticklevel = other.ticklevel
        if self.labellevel is None or (other.labellevel is not None and other.labellevel < self.labellevel):
            self.labellevel = other.labellevel


class timetexter:

    def __init__(self, format="%c"):
        self.format = format

    def labels(self, ticks):
        for tick in ticks:
            if tick.labellevel is not None and tick.label is None:
                tick.label = tick.strftime(self.format)


class timeaxis(axis._axis, _timemap):

     def __init__(self, parter=None, rater=rater.linear(), **args):
        axis._axis.__init__(self, divisor=None, **args)
        if self.fixmin and self.fixmax:
            self.relsize = self.max - self.min
            self.relsize = 0.000001*self.relsize.microseconds + self.relsize.seconds + 60*60*24.0*self.relsize.days
        self.parter = parter
        self.rater = rater





#!/usr/bin/env python
# -*- coding: ISO-8859-1 -*-
#
#
# Copyright (C) 2004 André Wobst <wobsta@users.sourceforge.net>
#
# This file is part of PyX (http://pyx.sourceforge.net/).
#
# PyX is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PyX is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyX; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


__all__ = ["data", "graph", "key", "style"]


for module in __all__:
    __import__(module, globals(), locals(), [])

from graph import graphxy



#!/usr/bin/env python
import sys, imp, re
sys.path[:0] = [".."]
import pyx
from pyx import *

text.set(mode="latex")
text.preamble(r"\renewcommand{\familydefault}{\ttdefault}")

c = canvas.canvas()

# data to be plotted
pf = graph.data.paramfunction("k", 0, 1, "color, xmin, xmax, ymin, ymax= k, k, 1, 0, 1")

# positioning is quite ugly ... but it works at the moment
y = 0
dy = -0.65

# we could use palette.__dict__ to get the instances, but we
# would loose the ordering ... instead we just parse the file:
p = re.compile("(?P<id>palette\\.(?P<name>[a-z]+)) += palette\\(.*\\)\n", re.IGNORECASE)
lines = imp.find_module("color", pyx.__path__)[0].readlines()
first = 1
for line in lines: # we yet don't use a file iterator
    m = p.match(line)
    if m:
        xaxis = graph.axis.lin(
                parter=graph.axis.parter.lin(tickdist=("0.5","0.1"), labeldist="1"),
                painter=graph.axis.painter.regular(innerticklength=None, labelattrs=None))
        if first:
            x2axis=graph.axis.linked(xaxis,
                painter=graph.axis.painter.linked(
                    innerticklength=None,
                    outerticklength=graph.axis.painter.ticklength.normal,
                    labelattrs=[]))
            first = 0
        else:
            x2axis=graph.axis.linked(xaxis, painter=graph.axis.painter.linked(innerticklength=None))
        g = c.insert(graph.graphxy(ypos=y, width=10, height=0.5, x=xaxis, x2=x2axis, y=graph.axis.lin(parter=None)))
        g.plot(pf, graph.style.rect(getattr(pyx.color.palette, m.group("name"))))
        g.dodata()
        g.finish()
        c.text(10.2, y + 0.15, m.group("id"), [text.size.footnotesize])
        y += dy


c.writeEPSfile("palettename", paperformat="a4")



#!/usr/bin/env python
# -*- coding: ISO-8859-1 -*-
#
#
# Copyright (C) 2002-2004 Jörg Lehmann <joergl@users.sourceforge.net>
# Copyright (C) 2002-2004 André Wobst <wobsta@users.sourceforge.net>
#
# This file is part of PyX (http://pyx.sourceforge.net/).
#
# PyX is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PyX is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyX; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# XXX what are the correct base classes of clip and pattern

"""The canvas module provides a PostScript canvas class and related classes

A canvas holds a collection of all elements that should be displayed together
with their attributes.
"""

import sys, cStringIO, time
import attr, base,  deco, unit, prolog, style, trafo, version

# known paperformats as tuple (width, height)

_paperformats = { "A4"      : ("210 t mm",  "297 t mm"),
                  "A3"      : ("297 t mm",  "420 t mm"),
                  "A2"      : ("420 t mm",  "594 t mm"),
                  "A1"      : ("594 t mm",  "840 t mm"),
                  "A0"      : ("840 t mm", "1188 t mm"),
                  "A0B"     : ("910 t mm", "1370 t mm"),
                  "LETTER"  : ("8.5 t inch", "11 t inch"),
                  "LEGAL"   : ("8.5 t inch", "14 t inch")}

#
# clipping class
#

class clip(base.PSCmd):

    """class for use in canvas constructor which clips to a path"""

    def __init__(self, path):
        """construct a clip instance for a given path"""
        self.path = path

    def bbox(self):
        # as a PSCmd a clipping path has NO influence on the bbox...
        return None

    def clipbbox(self):
        # ... but for clipping, we nevertheless need the bbox
        return self.path.bbox()

    def outputPS(self, file):
        file.write("newpath\n")
        self.path.outputPS(file)
        file.write("clip\n")

    def outputPDF(self, file):
        self.path.outputPDF(file)
        file.write("W n\n")

#
# general canvas class
#

class _canvas(base.PSCmd):

    """a canvas is a collection of PSCmds together with PSAttrs"""

    def __init__(self, attrs=[], texrunner=None):

        """construct a canvas

        The canvas can be modfied by supplying args, which have
        to be instances of one of the following classes:
         - trafo.trafo (leading to a global transformation of the canvas)
         - canvas.clip (clips the canvas)
         - base.PathStyle (sets some global attributes of the canvas)

        Note that, while the first two properties are fixed for the
        whole canvas, the last one can be changed via canvas.set().

        The texrunner instance used for the text method can be specified
        using the texrunner argument. It defaults to text.defaulttexrunner

        """

        self.PSOps     = []
        self.trafo     = trafo.trafo()
        self.clipbbox  = None
        if texrunner is not None:
            self.texrunner = texrunner
        else:
            # prevent cyclic imports
            import text
            self.texrunner = text.defaulttexrunner

        for attr in attrs:
            if isinstance(attr, trafo.trafo_pt):
                self.trafo = self.trafo*attr
                self.PSOps.append(attr)
            elif isinstance(attr, clip):
                if self.clipbbox is None:
                    self.clipbbox = attr.clipbbox().transformed(self.trafo)
                else:
                    self.clippbox *= attr.clipbbox().transformed(self.trafo)
                self.PSOps.append(attr)
            else:
                self.set([attr])

    def bbox(self):
        """returns bounding box of canvas"""
        obbox = None
        for cmd in self.PSOps:
            if isinstance(cmd, base.PSCmd):
                abbox = cmd.bbox()
                if obbox is None:
                    obbox = abbox
                elif abbox is not None:
                    obbox += abbox

        # transform according to our global transformation and
        # intersect with clipping bounding box (which have already been
        # transformed in canvas.__init__())
        if obbox is not None and self.clipbbox is not None:
            return obbox.transformed(self.trafo)*self.clipbbox
        elif obbox is not None:
            return obbox.transformed(self.trafo)
        else:
            return self.clipbbox

    def prolog(self):
        result = []
        for cmd in self.PSOps:
            result.extend(cmd.prolog())
        return result

    def outputPS(self, file):
        if self.PSOps:
            file.write("gsave\n")
            for cmd in self.PSOps:
                cmd.outputPS(file)
            file.write("grestore\n")

    def outputPDF(self, file):
        if self.PSOps:
            file.write("q\n") # gsave
            for cmd in self.PSOps:
                cmd.outputPDF(file)
            file.write("Q\n") # grestore

    def insert(self, PSOp, attrs=[]):
        """insert PSOp in the canvas.

        If args are given, then insert a canvas containing PSOp applying args.

        returns the PSOp

        """

        # XXX check for PSOp

        if attrs:
            sc = _canvas(attrs)
            sc.insert(PSOp)
            self.PSOps.append(sc)
        else:
            self.PSOps.append(PSOp)

        return PSOp

    def set(self, attrs):
        """sets styles args globally for the rest of the canvas
        """

        attr.checkattrs(attrs, [style.strokestyle, style.fillstyle])
        for astyle in attrs:
            self.insert(astyle)

    def draw(self, path, attrs):
        """draw path on canvas using the style given by args

        The argument attrs consists of PathStyles, which modify
        the appearance of the path, PathDecos, which add some new
        visual elements to the path, or trafos, which are applied
        before drawing the path.

        """

        attrs = attr.mergeattrs(attrs)
        attr.checkattrs(attrs, [deco.deco, style.fillstyle, style.strokestyle, trafo.trafo_pt])

        for t in attr.getattrs(attrs, [trafo.trafo_pt]):
            path = path.transformed(t)

        dp = deco.decoratedpath(path)

        # set global styles
        dp.styles = attr.getattrs(attrs, [style.fillstyle, style.strokestyle])

        # add path decorations and modify path accordingly
        for adeco in attr.getattrs(attrs, [deco.deco]):
            dp = adeco.decorate(dp)

        self.insert(dp)

    def stroke(self, path, attrs=[]):
        """stroke path on canvas using the style given by args

        The argument attrs consists of PathStyles, which modify
        the appearance of the path, PathDecos, which add some new
        visual elements to the path, or trafos, which are applied
        before drawing the path.

        """

        self.draw(path, [deco.stroked]+list(attrs))

    def fill(self, path, attrs=[]):
        """fill path on canvas using the style given by args

        The argument attrs consists of PathStyles, which modify
        the appearance of the path, PathDecos, which add some new
        visual elements to the path, or trafos, which are applied
        before drawing the path.

        """

        self.draw(path, [deco.filled]+list(attrs))

    def settexrunner(self, texrunner):
        """sets the texrunner to be used to within the text and text_pt methods"""

        self.texrunner = texrunner

    def text(self, x, y, atext, *args, **kwargs):
        """insert a text into the canvas

        inserts a textbox created by self.texrunner.text into the canvas

        returns the inserted textbox"""

        return self.insert(self.texrunner.text(x, y, atext, *args, **kwargs))


    def text_pt(self, x, y, atext, *args):
        """insert a text into the canvas

        inserts a textbox created by self.texrunner.text_pt into the canvas

        returns the inserted textbox"""

        return self.insert(self.texrunner.text_pt(x, y, atext, *args))

#
# canvas for patterns
#

class pattern(_canvas, attr.exclusiveattr, style.fillstyle):

    def __init__(self, painttype=1, tilingtype=1, xstep=None, ystep=None, bbox=None, trafo=None):
        attr.exclusiveattr.__init__(self, pattern)
        _canvas.__init__(self)
        attr.exclusiveattr.__init__(self, pattern)
        self.id = "pattern%d" % id(self)
        if painttype not in (1,2):
            raise ValueError("painttype must be 1 or 2")
        self.painttype = painttype
        if tilingtype not in (1,2,3):
            raise ValueError("tilingtype must be 1, 2 or 3")
        self.tilingtype = tilingtype
        self.xstep = xstep
        self.ystep = ystep
        self.patternbbox = bbox
        self.patterntrafo = trafo

    def bbox(self):
        return None

    def outputPS(self, file):
        file.write("%s setpattern\n" % self.id)

    def prolog(self):
        realpatternbbox = _canvas.bbox(self)
        if self.xstep is None:
           xstep = unit.topt(realpatternbbox.width())
        else:
           xstep = unit.topt(unit.length(self.xstep))
        if self.ystep is None:
            ystep = unit.topt(realpatternbbox.height())
        else:
           ystep = unit.topt(unit.length(self.ystep))
        if not xstep:
            raise ValueError("xstep in pattern cannot be zero")
        if not ystep:
            raise ValueError("ystep in pattern cannot be zero")
        patternbbox = self.patternbbox or realpatternbbox.enlarged("5 pt")

        patternprefix = "\n".join(("<<",
                                   "/PatternType 1",
                                   "/PaintType %d" % self.painttype,
                                   "/TilingType %d" % self.tilingtype,
                                   "/BBox[%s]" % str(patternbbox),
                                   "/XStep %g" % xstep,
                                   "/YStep %g" % ystep,
                                   "/PaintProc {\nbegin\n"))
        stringfile = cStringIO.StringIO()
        _canvas.outputPS(self, stringfile)
        patternproc = stringfile.getvalue()
        stringfile.close()
        patterntrafostring = self.patterntrafo is None and "matrix" or str(self.patterntrafo)
        patternsuffix = "end\n} bind\n>>\n%s\nmakepattern" % patterntrafostring

        pr = _canvas.prolog(self)
        pr.append(prolog.definition(self.id, "".join((patternprefix, patternproc, patternsuffix))))
        return pr

pattern.clear = attr.clearclass(pattern)

#
# The main canvas class
#

class canvas(_canvas):

    """a canvas is a collection of PSCmds together with PSAttrs"""

    def writeEPSfile(self, filename, paperformat=None, rotated=0, fittosize=0, margin="1 t cm",
                    bbox=None, bboxenlarge="1 t pt"):
        """write canvas to EPS file

        If paperformat is set to a known paperformat, the output will be centered on
        the page.

        If rotated is set, the output will first be rotated by 90 degrees.

        If fittosize is set, then the output is scaled to the size of the
        page (minus margin). In that case, the paperformat the specification
        of the paperformat is obligatory.

        The bbox parameter overrides the automatic bounding box determination.
        bboxenlarge may be used to enlarge the bbox of the canvas (or the
        manually specified bbox).
        """

        if filename[-4:]!=".eps":
            filename = filename + ".eps"

        try:
            file = open(filename, "w")
        except IOError:
            raise IOError("cannot open output file")

        abbox = bbox is not None and bbox or self.bbox()
        abbox = abbox.enlarged(bboxenlarge)
        ctrafo = None     # global transformation of canvas

        if rotated:
            ctrafo = trafo.rotate_pt(90,
                                     0.5*(abbox.llx+abbox.urx),
                                     0.5*(abbox.lly+abbox.ury))

        if paperformat:
            # center (optionally rotated) output on page
            try:
                width, height = _paperformats[paperformat.upper()]
            except KeyError:
                raise KeyError, "unknown paperformat '%s'" % paperformat
            width = unit.topt(width)
            height = unit.topt(height)

            if not ctrafo: ctrafo=trafo.trafo()

            ctrafo = ctrafo.translated_pt(0.5*(width -(abbox.urx-abbox.llx))-
                                       abbox.llx,
                                       0.5*(height-(abbox.ury-abbox.lly))-
                                       abbox.lly)

            if fittosize:
                # scale output to pagesize - margins
                margin = unit.topt(margin)
                if 2*margin > min(width, height):
                    raise RuntimeError("Margins too broad for selected paperformat. Aborting.")

                if rotated:
                    sfactor = min((height-2*margin)/(abbox.urx-abbox.llx),
                                  (width-2*margin)/(abbox.ury-abbox.lly))
                else:
                    sfactor = min((width-2*margin)/(abbox.urx-abbox.llx),
                                  (height-2*margin)/(abbox.ury-abbox.lly))

                ctrafo = ctrafo.scaled_pt(sfactor, sfactor, 0.5*width, 0.5*height)


        elif fittosize:
            raise ValueError("must specify paper size for fittosize")

        # if there has been a global transformation, adjust the bounding box
        # accordingly
        if ctrafo: abbox = abbox.transformed(ctrafo)

        file.write("%!PS-Adobe-3.0 EPSF 3.0\n")
        abbox.outputPS(file)
        file.write("%%%%Creator: PyX %s\n" % version.version)
        file.write("%%%%Title: %s\n" % filename)
        file.write("%%%%CreationDate: %s\n" %
                   time.asctime(time.localtime(time.time())))
        file.write("%%EndComments\n")

        file.write("%%BeginProlog\n")

        mergedprolog = []

        for pritem in self.prolog():
            for mpritem in mergedprolog:
                if mpritem.merge(pritem) is None: break
            else:
                mergedprolog.append(pritem)

        for pritem in mergedprolog:
            pritem.outputPS(file)

        file.write("%%EndProlog\n")

        # again, if there has occured global transformation, apply it now
        if ctrafo: ctrafo.outputPS(file)

        file.write("%f setlinewidth\n" % unit.topt(style.linewidth.normal))

        # here comes the actual content
        self.outputPS(file)

        file.write("showpage\n")
        file.write("%%Trailer\n")
        file.write("%%EOF\n")

    def writePDFfile(self, filename, bbox=None, bboxenlarge="1 t pt"):
        sys.stderr.write("*** PyX Warning: writePDFfile is experimental and supports only a subset of PyX's features\n")

        if filename[-4:]!=".pdf":
            filename = filename + ".pdf"

        try:
            file = open(filename, "wb")
        except IOError:
            raise IOError("cannot open output file")

        abbox = bbox is not None and bbox or self.bbox()
        abbox = abbox.enlarged(bboxenlarge)

        file.write("%%PDF-1.4\n%%%s%s%s%s\n" % (chr(195), chr(182), chr(195), chr(169)))
        reflist = [file.tell()]
        file.write("1 0 obj\n"
                   "<<\n"
                   "/Type /Catalog\n"
                   "/Outlines 2 0 R\n"
                   "/Pages 3 0 R\n"
                   ">>\n"
                   "endobj\n")
        reflist.append(file.tell())
        file.write("2 0 obj\n"
                   "<<\n"
                   "/Type Outlines\n"
                   "/Count 0\n"
                   ">>\n"
                   "endobj\n")
        reflist.append(file.tell())
        file.write("3 0 obj\n"
                   "<<\n"
                   "/Type /Pages\n"
                   "/Kids [4 0 R]\n"
                   "/Count 1\n"
                   ">>\n"
                   "endobj\n")
        reflist.append(file.tell())
        file.write("4 0 obj\n"
                   "<<\n"
                   "/Type /Page\n"
                   "/Parent 3 0 R\n")
        abbox.outputPDF(file)
        file.write("/Contents 5 0 R\n"
                   "/Resources << /ProcSet 7 0 R >>\n"
                   ">>\n"
                   "endobj\n")
        reflist.append(file.tell())
        file.write("5 0 obj\n"
                   "<< /Length 6 0 R >>\n"
                   "stream\n")
        streamstartpos = file.tell()
        style.linewidth.normal.outputPDF(file)
        self.outputPDF(file)
        streamendpos = file.tell()
        file.write("endstream\n"
                   "endobj\n")
        reflist.append(file.tell())
        file.write("6 0 obj\n"
                   "%s\n"
                   "endobj\n" % (streamendpos - streamstartpos))
        reflist.append(file.tell())
        file.write("7 0 obj\n"
                   "[/PDF]\n"
                   "endobj\n")
        xrefpos = file.tell()
        file.write("xref\n"
                   "0 8\n")
        file.write("0000000000 65535 f \n")
        for ref in reflist:
            file.write("%010i 00000 n \n" % ref)
        file.write("trailer\n"
                   "<<\n"
                   "/Size 8\n"
                   "/Root 1 0 R\n"
                   ">>\n"
                   "startxref\n"
                   "%i\n"
                   "%%%%EOF\n" % xrefpos)

    def writetofile(self, filename, *args, **kwargs):
        if filename[-4:] == ".eps":
            self.writeEPSfile(filename, *args, **kwargs)
        elif filename[-4:] == ".pdf":
            self.writePDFfile(filename, *args, **kwargs)
        else:
            sys.stderr.write("*** PyX Warning: deprecated usage of writetofile -- writetofile needs a filename extension or use an explicit call to writeEPSfile or the like\n")
            self.writeEPSfile(filename, *args, **kwargs)



#!/usr/bin/env python
# -*- coding: ISO-8859-1 -*-
#
#
# Copyright (C) 2002-2004 Jörg Lehmann <joergl@users.sourceforge.net>
# Copyright (C) 2002-2004 André Wobst <wobsta@users.sourceforge.net>
#
# This file is part of PyX (http://pyx.sourceforge.net/).
#
# PyX is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PyX is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyX; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import math
import attr, base, unit

# some helper routines

def _rmatrix(angle):
    phi = math.pi*angle/180.0

    return  ((math.cos(phi), -math.sin(phi)), 
             (math.sin(phi),  math.cos(phi)))

def _rvector(angle, x, y):
    phi = math.pi*angle/180.0

    return  ((1-math.cos(phi))*x + math.sin(phi)    *y,
              -math.sin(phi)   *x + (1-math.cos(phi))*y)


def _mmatrix(angle):
    phi = math.pi*angle/180.0

    return ( (math.cos(phi)*math.cos(phi)-math.sin(phi)*math.sin(phi),
              -2*math.sin(phi)*math.cos(phi)                ),
             (-2*math.sin(phi)*math.cos(phi),
              math.sin(phi)*math.sin(phi)-math.cos(phi)*math.cos(phi) ) )

def _det(matrix):
    return matrix[0][0]*matrix[1][1] - matrix[0][1]*matrix[1][0]

# Exception

class UndefinedResultError(ArithmeticError):
    pass

# trafo: affine transformations

class trafo_pt(base.PSOp, attr.attr):

    """affine transformation (coordinates in constructor in pts)

    Note that though the coordinates in the constructor are in
    pts (which is useful for internal purposes), all other
    methods only accept units in the standard user notation.

    """

    def __init__(self, matrix=((1,0),(0,1)), vector=(0,0)):
        if _det(matrix)==0:
            raise UndefinedResultError, "transformation matrix must not be singular" 
        else:
            self.matrix=matrix
        self.vector = vector

    def __mul__(self, other):
        if isinstance(other, trafo_pt):
            matrix = ( ( self.matrix[0][0]*other.matrix[0][0] +
                         self.matrix[0][1]*other.matrix[1][0],
                         self.matrix[0][0]*other.matrix[0][1] +
                         self.matrix[0][1]*other.matrix[1][1] ),
                       ( self.matrix[1][0]*other.matrix[0][0] +
                         self.matrix[1][1]*other.matrix[1][0],
                         self.matrix[1][0]*other.matrix[0][1] +
                         self.matrix[1][1]*other.matrix[1][1] )
                     )

            vector = ( self.matrix[0][0]*other.vector[0] +
                       self.matrix[0][1]*other.vector[1] +
                       self.vector[0],
                       self.matrix[1][0]*other.vector[0] +
                       self.matrix[1][1]*other.vector[1] +
                       self.vector[1] )

            return trafo_pt(matrix=matrix, vector=vector)
        else:
            raise NotImplementedError, "can only multiply two transformations"

    def __str__(self):
        return "[%f %f %f %f %f %f]" % \
               ( self.matrix[0][0], self.matrix[1][0],
                 self.matrix[0][1], self.matrix[1][1],
                 self.vector[0], self.vector[1] )

    def outputPS(self, file):
        file.write("[%f %f %f %f %f %f] concat\n" % \
                    ( self.matrix[0][0], self.matrix[1][0],
                      self.matrix[0][1], self.matrix[1][1],
                      self.vector[0], self.vector[1] ) )

    def outputPDF(self, file):
        file.write("%f %f %f %f %f %f cm\n" % \
                    ( self.matrix[0][0], self.matrix[1][0],
                      self.matrix[0][1], self.matrix[1][1],
                      self.vector[0], self.vector[1] ) )

    def bbox(self):
        return None

    def _apply(self, x, y):
        """apply transformation to point (x,y) (coordinates in pts)"""
        return (self.matrix[0][0]*x +
                self.matrix[0][1]*y +
                self.vector[0],
                self.matrix[1][0]*x +
                self.matrix[1][1]*y +
                self.vector[1])

    def apply(self, x, y):
        # before the transformation, we first have to convert to
        # our internal unit (i.e. pts)
        tx, ty = self._apply(unit.topt(x), unit.topt(y))

        # the end result can be converted back to general lengths
        return (unit.t_pt(tx), unit.t_pt(ty))

    def inverse(self):
        det = _det(self.matrix)                       # shouldn't be zero, but
        try:
          matrix = ( ( self.matrix[1][1]/det, -self.matrix[0][1]/det),
                     (-self.matrix[1][0]/det,  self.matrix[0][0]/det)
                   )
        except ZeroDivisionError:
           raise UndefinedResultError, "transformation matrix must not be singular" 
        return trafo_pt(matrix=matrix) * \
               trafo_pt(vector=(-self.vector[0], -self.vector[1]))

    def mirrored(self, angle):
        return mirror(angle)*self

    def rotated_pt(self, angle, x=None, y=None):
        return rotate_pt(angle, x, y)*self

    def rotated(self, angle, x=None, y=None):
        return rotate(angle, x, y)*self

    def scaled_pt(self, sx, sy=None, x=None, y=None):
        return scale_pt(sx, sy, x, y)*self

    def scaled(self, sx, sy=None, x=None, y=None):
        return scale(sx, sy, x, y)*self

    def slanted_pt(self, a, angle=0, x=None, y=None):
        return slant_pt(a, angle, x, y)*self

    def slanted(self, a, angle=0, x=None, y=None):
        return slant(a, angle, x, y)*self

    def translated_pt(self, x, y):
        return translate_pt(x,y)*self

    def translated(self, x, y):
        return translate(x, y)*self


class trafo(trafo_pt):

    """affine transformation"""

    def __init__(self, matrix=((1,0),(0,1)), vector=(0,0)):
        trafo_pt.__init__(self,
                        matrix,
                        (unit.topt(vector[0]), unit.topt(vector[1])))


#
# some standard transformations 
#

class mirror(trafo):
    def __init__(self,angle=0):
        trafo.__init__(self, matrix=_mmatrix(angle))


class rotate_pt(trafo_pt):
    def __init__(self, angle, x=None, y=None):
        vector = 0, 0
        if x is not None or y is not None:
            if x is None or y is None:
                raise (UndefinedResultError, 
                       "either specify both x and y or none of them")
            vector=_rvector(angle, x, y)

        trafo_pt.__init__(self,
                       matrix=_rmatrix(angle),
                       vector=vector)


class rotate(trafo_pt):
    def __init__(self, angle, x=None, y=None):
        vector = 0, 0 
        if x is not None or y is not None:
            if x is None or y is None:
                raise (UndefinedResultError, 
                       "either specify both x and y or none of them")
            vector=_rvector(angle, unit.topt(x), unit.topt(y))

        trafo_pt.__init__(self,
                       matrix=_rmatrix(angle),
                       vector=vector)


class scale_pt(trafo_pt):
    def __init__(self, sx, sy=None, x=None, y=None):
        sy = sy or sx
        if not sx or not sy:
            raise (UndefinedResultError, 
                   "one scaling factor is 0")
        vector = 0, 0 
        if x is not None or y is not None:
            if x is None or y is None:
                raise (UndefinedResultError, 
                       "either specify both x and y or none of them")
            vector = (1-sx)*x, (1-sy)*y

        trafo_pt.__init__(self, matrix=((sx,0),(0,sy)), vector=vector)


class scale(trafo):
    def __init__(self, sx, sy=None, x=None, y=None):
        sy = sy or sx
        if not sx or not sy:
            raise (UndefinedResultError, 
                   "one scaling factor is 0")
        vector = 0, 0 
        if x is not None or y is not None:
            if x is None or y is None:
                raise (UndefinedResultError, 
                       "either specify both x and y or none of them")
            vector = (1-sx)*x, (1-sy)*y

        trafo.__init__(self, matrix=((sx,0),(0,sy)), vector=vector)


class slant_pt(trafo_pt):
    def __init__(self, a, angle=0, x=None, y=None):
        t = ( rotate_pt(-angle, x, y)*
              trafo(matrix=((1, a), (0, 1)))*
              rotate_pt(angle, x, y) )
        trafo_pt.__init__(self, t.matrix, t.vector)


class slant(trafo):
    def __init__(self, a, angle=0, x=None, y=None):
        t = ( rotate(-angle, x, y)*
              trafo(matrix=((1, a), (0, 1)))*
              rotate(angle, x, y) )
        trafo.__init__(self, t.matrix, t.vector)


class translate_pt(trafo_pt):
    def __init__(self, x, y):
        trafo_pt.__init__(self, vector=(x, y))


class translate(trafo):
    def __init__(self, x, y):
        trafo.__init__(self, vector=(x, y))



# Partitioners (in the code the short form parter is used) take
# care of calculating appropriate tick positions for a given axis
# range. Automatic partitioners create several tick lists, which are
# rated by an axis rater instance afterwards while manual partitioners
# create a single tick list only, which thus doesn't need to be rated.
#
# Note that the partitioning uses fractional number arithmetics. For
# that, tick instances can be initialized with floats using a fixed
# precision but also with strings as shown.

import math
from pyx import *
from pyx.graph import axis

p = path.curve(0, 0, 3, 0, 1, 4, 4, 4)

myparter = axis.parter.linear(["1/3", "1/6"])

c = canvas.canvas()
c.insert(axis.pathaxis(p, axis.linear(min=0, max=1, parter=myparter)))
c.insert(axis.pathaxis(p.transformed(trafo.translate(4, 0)),
                       axis.linear(min=0, max=1, parter=myparter,
                                   texter=axis.texter.rational())))
c.writeEPSfile("parter")



from pyx import *

# set properties of the defaulttexrunner, e.g. switch to LaTeX
text.set(mode="latex")

c = canvas.canvas()
# the canvas, by default, uses the defaulttexrunner from the text module
# (this can be changed by the canvas method settexrunner)
c.text(0, 0, r"This is \LaTeX.")

# you can have several texrunners (several running TeX/LaTeX instances)
plaintex = text.texrunner() # plain TeX instance
c.insert(plaintex.text(0, -1, r"This is plain \TeX."))

c.writeEPSfile("latex")



#!/usr/bin/env python
# -*- coding: ISO-8859-1 -*-
#
#
# Copyright (C) 2002-2004 Jörg Lehmann <joergl@users.sourceforge.net>
# Copyright (C) 2002-2004 André Wobst <wobsta@users.sourceforge.net>
#
# This file is part of PyX (http://pyx.sourceforge.net/).
#
# PyX is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PyX is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyX; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


class nodefault: pass


def isstring(arg):
    "arg is string-like (cf. python cookbook 3.2)"
    try: arg + ''
    except: return 0
    return 1


def isnumber(arg):
    "arg is number-like"
    try: arg + 0
    except: return 0
    return 1


def isinteger(arg):
    "arg is integer-like"
    try:
        if type(arg + 0.0) is type(arg):
            return 0
        return 1
    except: return 0


def issequence(arg):
    """arg is sequence-like (e.g. has a len)
       a string is *not* considered to be a sequence"""
    if isstring(arg): return 0
    try: len(arg)
    except: return 0
    return 1


def ensuresequence(arg):
    """return arg or (arg,) depending on the result of issequence,
       None is converted to ()"""
    if isstring(arg): return (arg,)
    if arg is None: return ()
    if issequence(arg): return arg
    return (arg,)


def ensurelist(arg):
    """return list(arg) or [arg] depending on the result of isequence,
       None is converted to []"""
    if isstring(arg): return [arg]
    if arg is None: return []
    if issequence(arg): return list(arg)
    return [arg]

def getitemno(arg, n):
    """get item number n if arg is a sequence (when the sequence
       is not long enough, None is returned), otherweise arg is
       returned"""
    if issequence(arg):
        try: return arg[n]
        except: return None
    else:
        return arg


def issequenceofsequences(arg):
    """check if arg has a sequence or None as it's first entry"""
    return issequence(arg) and len(arg) and (issequence(arg[0]) or arg[0] is None)


def getsequenceno(arg, n):
    """get sequence number n if arg is a sequence of sequences (when
       the sequence is not long enough, None is returned), otherwise
       arg is returned"""
    if issequenceofsequences(arg):
        try: return arg[n]
        except: return None
    else:
        return arg



#!/usr/bin/env python
# -*- coding: ISO-8859-1 -*-
#
#
# Copyright (C) 2002-2004 Jörg Lehmann <joergl@users.sourceforge.net>
# Copyright (C) 2003-2004 Michael Schindler <m-schindler@users.sourceforge.net>
# Copyright (C) 2002-2004 André Wobst <wobsta@users.sourceforge.net>
#
# This file is part of PyX (http://pyx.sourceforge.net/).
#
# PyX is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PyX is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyX; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


import math
from pyx import attr, helper
from pyx.graph.axis import painter, parter, rater, texter, tick


class _Imap:
    """interface definition of a map
    maps convert a value into another value by bijective transformation f"""

    def convert(self, x):
        "returns f(x)"

    def invert(self, y):
        "returns f^-1(y) where f^-1 is the inverse transformation (x=f^-1(f(x)) for all x)"

    def setbasepoints(self, basepoints):
        """set basepoints for the convertions
        basepoints are tuples (x, y) with y == f(x) and x == f^-1(y)
        the number of basepoints needed might depend on the transformation
        usually two pairs are needed like for linear maps, logarithmic maps, etc."""


class _linmap:
    "linear mapping"

    __implements__ = _Imap

    def setbasepoints(self, basepoints):
        self.x1 = float(basepoints[0][0])
        self.y1 = float(basepoints[0][1])
        self.x2 = float(basepoints[1][0])
        self.y2 = float(basepoints[1][1])
        self.dydx = (self.y2 - self.y1) / (self.x2 - self.x1)
        self.dxdy = (self.x2 - self.x1) / (self.y2 - self.y1)

    def convert(self, value):
        return self.y1 + self.dydx * (float(value) - self.x1)

    def invert(self, value):
        return self.x1 + self.dxdy * (float(value) - self.y1)


class _logmap:
    "logarithmic mapping"
    __implements__ = _Imap

    def setbasepoints(self, basepoints):
        self.x1 = float(math.log(basepoints[0][0]))
        self.y1 = float(basepoints[0][1])
        self.x2 = float(math.log(basepoints[1][0]))
        self.y2 = float(basepoints[1][1])
        self.dydx = (self.y2 - self.y1) / (self.x2 - self.x1)
        self.dxdy = (self.x2 - self.x1) / (self.y2 - self.y1)

    def convert(self, value):
        return self.y1 + self.dydx * (math.log(float(value)) - self.x1)

    def invert(self, value):
        return math.exp(self.x1 + self.dxdy * (float(value) - self.y1))


class _Iaxis:
    """interface definition of a axis
    - an axis should implement an convert and invert method like
      _Imap, but this is not part of this interface definition;
      one possibility is to mix-in a proper map class, but special
      purpose axes might do something else
    - an axis has the instance variable axiscanvas after the finish
      method was called
    - an axis might have further instance variables (title, ticks)
      to be used in combination with appropriate axispainters"""

    def convert(self, x):
        "convert a value into graph coordinates"

    def invert(self, v):
        "invert a graph coordinate to a axis value"

    def getrelsize(self):
        """returns the relative size (width) of the axis
        - for use in split axis, bar axis etc.
        - might return None if no size is available"""

    # TODO: describe adjustrange
    def setrange(self, min=None, max=None):
        """set the axis data range
        - the type of min and max must fit to the axis
        - min<max; the axis might be reversed, but this is
          expressed internally only (min<max all the time)
        - the axis might not apply the change of the range
          (e.g. when the axis range is fixed by the user),
          but usually the range is extended to contain the
          given range
        - for invalid parameters (e.g. negativ values at an
          logarithmic axis), an exception should be raised
        - a RuntimeError is raised, when setrange is called
          after the finish method"""

    def getrange(self):
        """return data range as a tuple (min, max)
        - min<max; the axis might be reversed, but this is
          expressed internally only
        - a RuntimeError exception is raised when no
          range is available"""

    def finish(self, axispos):
        """finishes the axis
        - axispos implements _Iaxispos
        - sets the instance axiscanvas, which is insertable into the
          graph to finally paint the axis
        - any modification of the axis range should be disabled after
          the finish method was called"""
        # TODO: be more specific about exceptions

    def createlinkaxis(self, **kwargs):
        """create a link axis to the axis itself
        - typically, a link axis is a axis, which share almost
          all properties with the axis it is linked to
        - typically, the painter gets replaced by a painter
          which doesn't put any text to the axis"""


class _axis:
    """base implementation a regular axis
    - typical usage is to mix-in a linmap or a logmap to
      complete the axis interface
    - note that some methods of this class want to access a
      parter and a rater; those attributes implementing _Iparter
      and _Irater should be initialized by the constructors
      of derived classes"""

    def __init__(self, min=None, max=None, reverse=0, divisor=None,
                       title=None, painter=painter.regular(), texter=texter.mixed(),
                       density=1, maxworse=2, manualticks=[]):
        """initializes the instance
        - min and max fix the axis minimum and maximum, respectively;
          they are determined by the data to be plotted, when not fixed
        - reverse (boolean) reverses the minimum and the maximum of
          the axis
        - numerical divisor for the axis partitioning
        - title is a string containing the axis title
        - axispainter is the axis painter (should implement _Ipainter)
        - texter is the texter (should implement _Itexter)
        - density is a global parameter for the axis paritioning and
          axis rating; its default is 1, but the range 0.5 to 2.5 should
          be usefull to get less or more ticks by the automatic axis
          partitioning
        - maxworse is a number of trials with worse tick rating
          before giving up (usually it should not be needed to increase
          this value; increasing the number will slow down the automatic
          axis partitioning considerably)
        - manualticks and the partitioner results are mixed
          by mergeticklists
        - note that some methods of this class want to access a
          parter and a rater; those attributes implementing _Iparter
          and _Irater should be initialized by the constructors
          of derived classes"""
        if min is not None and max is not None and min > max:
            min, max, reverse = max, min, not reverse
        self.fixmin, self.fixmax, self.min, self.max, self.reverse = min is not None, max is not None, min, max, reverse
        self.divisor = divisor
        self.usedivisor = 0
        self.title = title
        self.painter = painter
        self.texter = texter
        self.density = density
        self.maxworse = maxworse
        self.manualticks = self.checkfraclist(manualticks)
        self.axiscanvas = None
        self._setrange()

    def _setrange(self, min=None, max=None):
        if not self.fixmin and min is not None and (self.min is None or min < self.min):
            self.min = min
        if not self.fixmax and max is not None and (self.max is None or max > self.max):
            self.max = max
        if None not in (self.min, self.max) and self.min != self.max:
            if self.reverse:
                if self.usedivisor and self.divisor is not None:
                    self.setbasepoints(((self.min/self.divisor, 1), (self.max/self.divisor, 0)))
                else:
                    self.setbasepoints(((self.min, 1), (self.max, 0)))
            else:
                if self.usedivisor and self.divisor is not None:
                    self.setbasepoints(((self.min/self.divisor, 0), (self.max/self.divisor, 1)))
                else:
                    self.setbasepoints(((self.min, 0), (self.max, 1)))

    def _getrange(self):
        return self.min, self.max

    def _forcerange(self, range):
        self.min, self.max = range
        self._setrange()

    def setrange(self, min=None, max=None):
        if self.usedivisor and self.divisor is not None:
            min = float(self.divisor) * self.divisor
            max = float(self.divisor) * self.divisor
        oldmin, oldmax = self.min, self.max
        self._setrange(min, max)
        if self.axiscanvas is not None and ((oldmin != self.min) or (oldmax != self.max)):
            raise RuntimeError("range modification while axis was already finished")

    zero = 0.0

    def adjustrange(self, points, index, deltaindex=None, deltaminindex=None, deltamaxindex=None):
        min = max = None
        if len([x for x in [deltaindex, deltaminindex, deltamaxindex] if x is not None]) > 1:
            raise RuntimeError("only one of delta???index should set")
        if deltaindex is not None:
            deltaminindex = deltamaxindex = deltaindex
        if deltaminindex is not None:
            for point in points:
                try:
                    value = point[index] - point[deltaminindex] + self.zero
                except:
                    pass
                else:
                    if min is None or value < min: min = value
                    if max is None or value > max: max = value
        elif deltamaxindex is not None:
            for point in points:
                try:
                    value = point[index] + point[deltamaxindex] + self.zero
                except:
                    pass
                else:
                    if min is None or value < min: min = value
                    if max is None or value > max: max = value
        else:
            for point in points:
                try:
                    value = point[index] + self.zero
                except:
                    pass
                else:
                    if min is None or value < min: min = value
                    if max is None or value > max: max = value
        self.setrange(min, max)

    def getrange(self):
        if self.min is not None and self.max is not None:
            if self.usedivisor and self.divisor is not None:
                return float(self.min) / self.divisor, float(self.max) / self.divisor
            else:
                return self.min, self.max

    def checkfraclist(self, fracs):
        "orders a list of fracs, equal entries are not allowed"
        if not len(fracs): return []
        sorted = list(fracs)
        sorted.sort()
        last = sorted[0]
        for item in sorted[1:]:
            if last == item:
                raise ValueError("duplicate entry found")
            last = item
        return sorted

    def finish(self, axispos):
        if self.axiscanvas is not None: return

        # temorarily enable the axis divisor
        self.usedivisor = 1
        self._setrange()

        # lesspart and morepart can be called after defaultpart;
        # this works although some axes may share their autoparting,
        # because the axes are processed sequentially
        first = 1
        if self.parter is not None:
            min, max = self.getrange()
            self.ticks = tick.mergeticklists(self.manualticks,
                                             self.parter.defaultpart(min, max, not self.fixmin, not self.fixmax))
            worse = 0
            nextpart = self.parter.lesspart
            while nextpart is not None:
                newticks = nextpart()
                if newticks is not None:
                    newticks = tick.mergeticklists(self.manualticks, newticks)
                    if first:
                        bestrate = self.rater.rateticks(self, self.ticks, self.density)
                        bestrate += self.rater.raterange(self.convert(self.ticks[-1])-
                                                         self.convert(self.ticks[0]), 1)
                        variants = [[bestrate, self.ticks]]
                        first = 0
                    newrate = self.rater.rateticks(self, newticks, self.density)
                    newrate += self.rater.raterange(self.convert(newticks[-1])-
                                                    self.convert(newticks[0]), 1)
                    variants.append([newrate, newticks])
                    if newrate < bestrate:
                        bestrate = newrate
                        worse = 0
                    else:
                        worse += 1
                else:
                    worse += 1
                if worse == self.maxworse and nextpart == self.parter.lesspart:
                    worse = 0
                    nextpart = self.parter.morepart
                if worse == self.maxworse and nextpart == self.parter.morepart:
                    nextpart = None
        else:
            self.ticks =self.manualticks

        # rating, when several choises are available
        if not first:
            variants.sort()
            if self.painter is not None:
                i = 0
                bestrate = None
                while i < len(variants) and (bestrate is None or variants[i][0] < bestrate):
                    saverange = self._getrange()
                    self.ticks = variants[i][1]
                    if len(self.ticks):
                        self.setrange(self.ticks[0], self.ticks[-1])
                    self.texter.labels(self.ticks)
                    ac = self.painter.paint(axispos, self)
                    ratelayout = self.rater.ratelayout(ac, self.density)
                    if ratelayout is not None:
                        variants[i][0] += ratelayout
                        variants[i].append(ac)
                    else:
                        variants[i][0] = None
                    if variants[i][0] is not None and (bestrate is None or variants[i][0] < bestrate):
                        bestrate = variants[i][0]
                    self._forcerange(saverange)
                    i += 1
                if bestrate is None:
                    raise RuntimeError("no valid axis partitioning found")
                variants = [variant for variant in variants[:i] if variant[0] is not None]
                variants.sort()
                self.ticks = variants[0][1]
                if len(self.ticks):
                    self.setrange(self.ticks[0], self.ticks[-1])
                self.axiscanvas = variants[0][2]
            else:
                self.ticks = variants[0][1]
                self.texter.labels(self.ticks)
                if len(self.ticks):
                    self.setrange(self.ticks[0], self.ticks[-1])
                self.axiscanvas = painter.axiscanvas()
        else:
            if len(self.ticks):
                self.setrange(self.ticks[0], self.ticks[-1])
            self.texter.labels(self.ticks)
            if self.painter is not None:
                self.axiscanvas = self.painter.paint(axispos, self)
            else:
                self.axiscanvas = painter.axiscanvas()

        # disable the axis divisor
        self.usedivisor = 0
        self._setrange()

    def createlinkaxis(self, **args):
        return linked(self, **args)


class linear(_axis, _linmap):
    """implementation of a linear axis"""

    __implements__ = _Iaxis

    def __init__(self, parter=parter.autolinear(), rater=rater.linear(), **args):
        """initializes the instance
        - the parter attribute implements _Iparter
        - manualticks and the partitioner results are mixed
          by mergeticklists
        - the rater implements _Irater and is used to rate different
          tick lists created by the partitioner (after merging with
          manully set ticks)
        - futher keyword arguments are passed to _axis"""
        _axis.__init__(self, **args)
        if self.fixmin and self.fixmax:
            self.relsize = self.max - self.min
        self.parter = parter
        self.rater = rater

lin = linear


class logarithmic(_axis, _logmap):
    """implementation of a logarithmic axis"""

    __implements__ = _Iaxis

    def __init__(self, parter=parter.autologarithmic(), rater=rater.logarithmic(), **args):
        """initializes the instance
        - the parter attribute implements _Iparter
        - manualticks and the partitioner results are mixed
          by mergeticklists
        - the rater implements _Irater and is used to rate different
          tick lists created by the partitioner (after merging with
          manully set ticks)
        - futher keyword arguments are passed to _axis"""
        _axis.__init__(self, **args)
        if self.fixmin and self.fixmax:
            self.relsize = math.log(self.max) - math.log(self.min)
        self.parter = parter
        self.rater = rater

log = logarithmic


class _linked:
    """base for a axis linked to an already existing axis
    - almost all properties of the axis are "copied" from the
      axis this axis is linked to
    - usually, linked axis are used to create an axis to an
      existing axis with different painting properties; linked
      axis can be used to plot an axis twice at the opposite
      sides of a graphxy or even to share an axis between
      different graphs!"""

    __implements__ = _Iaxis

    def __init__(self, linkedaxis, painter=painter.linked()):
        """initializes the instance
        - it gets a axis this axis is linked to
        - it gets a painter to be used for this linked axis"""
        self.linkedaxis = linkedaxis
        self.painter = painter
        self.axiscanvas = None

    def __getattr__(self, attr):
        """access to unkown attributes are handed over to the
        axis this axis is linked to"""
        return getattr(self.linkedaxis, attr)

    def finish(self, axispos):
        """finishes the axis
        - instead of performing the hole finish process
          (paritioning, rating, etc.) just a painter call
          is performed"""

        if self.axiscanvas is None:
            if self.linkedaxis.axiscanvas is None:
                raise RuntimeError("link axis finish method called before the finish method of the original axis")
            self.axiscanvas = self.painter.paint(axispos, self)


class linked(_linked):
    """a axis linked to an already existing regular axis
    - adds divisor handling to _linked"""

    __implements__ = _Iaxis

    def finish(self, axispos):
        # temporarily enable the linkedaxis divisor
        self.linkedaxis.usedivisor = 1
        self.linkedaxis._setrange()

        _linked.finish(self, axispos)

        # disable the linkedaxis divisor again
        self.linkedaxis.usedivisor = 0
        self.linkedaxis._setrange()


class split:
    """implementation of a split axis
    - a split axis contains several (sub-)axes with
      non-overlapping data ranges -- between these subaxes
      the axis is "splitted"
    - (just to get sure: a split axis can contain other
      split axes as its subaxes)"""

    __implements__ = _Iaxis, painter._Iaxispos

    def __init__(self, subaxes, splitlist=[0.5], splitdist=0.1, relsizesplitdist=1,
                       title=None, painter=painter.split()):
        """initializes the instance
        - subaxes is a list of subaxes
        - splitlist is a list of graph coordinates, where the splitting
          of the main axis should be performed; if the list isn't long enough
          for the subaxes, missing entries are considered to be None
        - splitdist is the size of the splitting in graph coordinates, when
          the associated splitlist entry is not None
        - relsizesplitdist: a None entry in splitlist means, that the
          position of the splitting should be calculated out of the
          relsize values of conrtibuting subaxes (the size of the
          splitting is relsizesplitdist in values of the relsize values
          of the axes)
        - title is the title of the axis as a string
        - painter is the painter of the axis; it should be specialized to
          the split axis
        - the relsize of the split axis is the sum of the relsizes of the
          subaxes including the relsizesplitdist"""
        self.subaxes = subaxes
        self.painter = painter
        self.title = title
        self.splitlist = splitlist
        for subaxis in self.subaxes:
            subaxis.vmin = None
            subaxis.vmax = None
        self.subaxes[0].vmin = 0
        self.subaxes[0].vminover = None
        self.subaxes[-1].vmax = 1
        self.subaxes[-1].vmaxover = None
        for i in xrange(len(self.splitlist)):
            if self.splitlist[i] is not None:
                self.subaxes[i].vmax = self.splitlist[i] - 0.5*splitdist
                self.subaxes[i].vmaxover = self.splitlist[i]
                self.subaxes[i+1].vmin = self.splitlist[i] + 0.5*splitdist
                self.subaxes[i+1].vminover = self.splitlist[i]
        i = 0
        while i < len(self.subaxes):
            if self.subaxes[i].vmax is None:
                j = relsize = relsize2 = 0
                while self.subaxes[i + j].vmax is None:
                    relsize += self.subaxes[i + j].relsize + relsizesplitdist
                    j += 1
                relsize += self.subaxes[i + j].relsize
                vleft = self.subaxes[i].vmin
                vright = self.subaxes[i + j].vmax
                for k in range(i, i + j):
                    relsize2 += self.subaxes[k].relsize
                    self.subaxes[k].vmax = vleft + (vright - vleft) * relsize2 / float(relsize)
                    relsize2 += 0.5 * relsizesplitdist
                    self.subaxes[k].vmaxover = self.subaxes[k + 1].vminover = vleft + (vright - vleft) * relsize2 / float(relsize)
                    relsize2 += 0.5 * relsizesplitdist
                    self.subaxes[k+1].vmin = vleft + (vright - vleft) * relsize2 / float(relsize)
                if i == 0 and i + j + 1 == len(self.subaxes):
                    self.relsize = relsize
                i += j + 1
            else:
                i += 1

        self.fixmin = self.subaxes[0].fixmin
        if self.fixmin:
            self.min = self.subaxes[0].min
        self.fixmax = self.subaxes[-1].fixmax
        if self.fixmax:
            self.max = self.subaxes[-1].max

        self.axiscanvas = None

    def getrange(self):
        min = self.subaxes[0].getrange()
        max = self.subaxes[-1].getrange()
        try:
            return min[0], max[1]
        except TypeError:
            return None

    def setrange(self, min, max):
        self.subaxes[0].setrange(min, None)
        self.subaxes[-1].setrange(None, max)

    def adjustrange(self, *args, **kwargs):
        self.subaxes[0].adjustrange(*args, **kwargs)
        self.subaxes[-1].adjustrange(*args, **kwargs)

    def convert(self, value):
        # TODO: proper raising exceptions (which exceptions go thru, which are handled before?)
        if value < self.subaxes[0].max:
            return self.subaxes[0].vmin + self.subaxes[0].convert(value)*(self.subaxes[0].vmax-self.subaxes[0].vmin)
        for axis in self.subaxes[1:-1]:
            if value > axis.min and value < axis.max:
                return axis.vmin + axis.convert(value)*(axis.vmax-axis.vmin)
        if value > self.subaxes[-1].min:
            return self.subaxes[-1].vmin + self.subaxes[-1].convert(value)*(self.subaxes[-1].vmax-self.subaxes[-1].vmin)
        raise ValueError("value couldn't be assigned to a split region")

    def finish(self, axispos):
        if self.axiscanvas is None:
            self.axiscanvas = self.painter.paint(axispos, self)

    def createlinkaxis(self, **args):
        return linkedsplit(self, **args)


class omitsubaxispainter: pass

class linkedsplit(_linked):
    """a split axis linked to an already existing split axis
    - inherits the access to a linked axis -- as before,
      basically only the painter is replaced
    - it takes care of the creation of linked axes of
      the subaxes"""

    __implements__ = _Iaxis

    def __init__(self, linkedaxis, painter=painter.linkedsplit(), subaxispainter=omitsubaxispainter):
        """initializes the instance
        - linkedaxis is the axis this axis becomes linked to
        - painter is axispainter instance for this linked axis
        - subaxispainter is a changeable painter to be used for linked
          subaxes; if omitsubaxispainter the createlinkaxis method of
          the subaxis are called without a painter parameter"""
        _linked.__init__(self, linkedaxis, painter=painter)
        self.subaxes = []
        for subaxis in linkedaxis.subaxes:
            painter = attr.selectattr(subaxispainter, len(self.subaxes), len(linkedaxis.subaxes))
            if painter is omitsubaxispainter:
                self.subaxes.append(subaxis.createlinkaxis())
            else:
                self.subaxes.append(subaxis.createlinkaxis(painter=painter))


class bar:
    """implementation of a axis for bar graphs
    - a bar axes is different from a split axis by the way it
      selects its subaxes: the convert method gets a list,
      where the first entry is a name selecting a subaxis out
      of a list; instead of the term "bar" or "subaxis" the term
      "item" will be used here
    - the bar axis stores a list of names be identify the items;
      the names might be of any time (strings, integers, etc.);
      the names can be printed as the titles for the items, but
      alternatively the names might be transformed by the texts
      dictionary, which maps a name to a text to be used to label
      the items in the painter
    - usually, there is only one subaxis, which is used as
      the subaxis for all items
    - alternatively it is also possible to use another bar axis
      as a multisubaxis; it is copied via the createsubaxis
      method whenever another subaxis is needed (by that a
      nested bar axis with a different number of subbars at
      each item can be created)
    - any axis can be a subaxis of a bar axis; if no subaxis
      is specified at all, the bar axis simulates a linear
      subaxis with a fixed range of 0 to 1"""

    def __init__(self, subaxis=None, multisubaxis=None,
                       dist=0.5, firstdist=None, lastdist=None,
                       names=None, texts={},
                       title=None, painter=painter.bar()):
        """initialize the instance
        - subaxis contains a axis to be used as the subaxis
          for all items
        - multisubaxis might contain another bar axis instance
          to be used to construct a new subaxis for each item;
          (by that a nested bar axis with a different number
          of subbars at each item can be created)
        - only one of subaxis or multisubaxis can be set; if neither
          of them is set, the bar axis behaves like having a linear axis
          as its subaxis with a fixed range 0 to 1
        - the title attribute contains the axis title as a string
        - the dist is a relsize to be used as the distance between
          the items
        - the firstdist and lastdist are the distance before the
          first and after the last item, respectively; when set
          to None (the default), 0.5*dist is used
        - the relsize of the bar axis is the sum of the
          relsizes including all distances between the items"""
        self.dist = dist
        if firstdist is not None:
            self.firstdist = firstdist
        else:
            self.firstdist = 0.5 * dist
        if lastdist is not None:
            self.lastdist = lastdist
        else:
            self.lastdist = 0.5 * dist
        self.relsizes = None
        self.multisubaxis = multisubaxis
        if self.multisubaxis is not None:
            if subaxis is not None:
                raise RuntimeError("either use subaxis or multisubaxis")
            self.subaxis = []
        else:
            self.subaxis = subaxis
        self.title = title
        self.painter = painter
        self.axiscanvas = None
        self.names = []

    def createsubaxis(self):
        return bar(subaxis=self.multisubaxis.subaxis,
                   multisubaxis=self.multisubaxis.multisubaxis,
                   title=self.multisubaxis.title,
                   dist=self.multisubaxis.dist,
                   firstdist=self.multisubaxis.firstdist,
                   lastdist=self.multisubaxis.lastdist,
                   painter=self.multisubaxis.painter)

    def getrange(self):
        # TODO: we do not yet have a proper range handling for a bar axis
        return None

    def setrange(self, min=None, max=None):
        # TODO: we do not yet have a proper range handling for a bar axis
        raise RuntimeError("range handling for a bar axis is not implemented")

    def setname(self, name, *subnames):
        """add a name to identify an item at the bar axis
        - by using subnames, nested name definitions are
          possible
        - a style (or the user itself) might use this to
          insert new items into a bar axis
        - setting self.relsizes to None forces later recalculation"""
        if name not in self.names:
            self.relsizes = None
            self.names.append(name)
            if self.multisubaxis is not None:
                self.subaxis.append(self.createsubaxis())
        if len(subnames):
            if self.multisubaxis is not None:
                if self.subaxis[self.names.index(name)].setname(*subnames):
                    self.relsizes = None
            else:
                if self.subaxis.setname(*subnames):
                    self.relsizes = None
        return self.relsizes is not None

    def adjustrange(self, points, index, subnames=None):
        if subnames is None:
            subnames = []
        for point in points:
            self.setname(point[index], *subnames)

    def updaterelsizes(self):
        # guess what it does: it recalculates relsize attribute
        self.relsizes = [i*self.dist + self.firstdist for i in range(len(self.names) + 1)]
        self.relsizes[-1] += self.lastdist - self.dist
        if self.multisubaxis is not None:
            subrelsize = 0
            for i in range(1, len(self.relsizes)):
                self.subaxis[i-1].updaterelsizes()
                subrelsize += self.subaxis[i-1].relsizes[-1]
                self.relsizes[i] += subrelsize
        else:
            if self.subaxis is None:
                subrelsize = 1
            else:
                self.subaxis.updaterelsizes()
                subrelsize = self.subaxis.relsizes[-1]
            for i in range(1, len(self.relsizes)):
                self.relsizes[i] += i * subrelsize

    def convert(self, value):
        """bar axis convert method
        - the value should be a list, where the first entry is
          a member of the names (set in the constructor or by the
          setname method); this first entry identifies an item in
          the bar axis
        - following values are passed to the appropriate subaxis
          convert method
        - when there is no subaxis, the convert method will behave
          like having a linear axis from 0 to 1 as subaxis"""
        # TODO: proper raising exceptions (which exceptions go thru, which are handled before?)
        if not self.relsizes:
            self.updaterelsizes()
        pos = self.names.index(value[0])
        if len(value) == 2:
            if self.subaxis is None:
                subvalue = value[1]
            else:
                if self.multisubaxis is not None:
                    subvalue = value[1] * self.subaxis[pos].relsizes[-1]
                else:
                    subvalue = value[1] * self.subaxis.relsizes[-1]
        else:
            if self.multisubaxis is not None:
                subvalue = self.subaxis[pos].convert(value[1:]) * self.subaxis[pos].relsizes[-1]
            else:
                subvalue = self.subaxis.convert(value[1:]) * self.subaxis.relsizes[-1]
        return (self.relsizes[pos] + subvalue) / float(self.relsizes[-1])

    def finish(self, axispos):
        if self.axiscanvas is None:
            if self.multisubaxis is not None:
                for name, subaxis in zip(self.names, self.subaxis):
                    subaxis.vmin = self.convert((name, 0))
                    subaxis.vmax = self.convert((name, 1))
            self.axiscanvas = self.painter.paint(axispos, self)

    def createlinkaxis(self, **args):
        return linkedbar(self, **args)


class linkedbar(_linked):
    """a bar axis linked to an already existing bar axis
    - inherits the access to a linked axis -- as before,
      basically only the painter is replaced
    - it must take care of the creation of linked axes of
      the subaxes"""

    __implements__ = _Iaxis

    def __init__(self, linkedaxis, painter=painter.linkedbar()):
        """initializes the instance
        - it gets a axis this linkaxis is linked to
        - it gets a painter to be used for this linked axis"""
        _linked.__init__(self, linkedaxis, painter=painter)
        if self.multisubaxis is not None:
            self.subaxis = [subaxis.createlinkaxis() for subaxis in self.linkedaxis.subaxis]
        elif self.subaxis is not None:
            self.subaxis = self.subaxis.createlinkaxis()


def pathaxis(path, axis, **kwargs):
    """creates an axiscanvas for an axis along a path"""
    axis.finish(painter.pathaxispos(path, axis.convert, **kwargs))
    return axis.axiscanvas



#!/usr/bin/env python
# -*- coding: ISO-8859-1 -*-
#
#
# Copyright (C) 2003-2004 Michael Schindler <m-schindler@users.sourceforge.net>
#
# This file is part of PyX (http://pyx.sourceforge.net/).
#
# PyX is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PyX is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyX; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


import math
from math import pi, sin, cos, atan2, tan, hypot, acos, sqrt
import path, trafo, unit, helper
try:
    from math import radians, degrees
except ImportError:
    # fallback implementation for Python 2.1 and below
    def radians(x): return x*pi/180
    def degrees(x): return x*180/pi


#########################
##   helpers
#########################

def _topt(length, default_type=None):
    if length is None: return None
    if default_type is not None:
        return unit.topt(unit.length(length, default_type=default_type))
    else:
        return unit.topt(unit.length(length))

class connector_pt(path.normpath):

    def omitends(self, box1, box2):
        """intersect a path with the boxes' paths"""

        # cut off the start of self
        # XXX how can decoration of this box1.path() be handled?
        sp = self.intersect(box1.path())[0]
        sp.sort()
        if sp:
            self.subpaths = self.split(sp[-1:])[1].subpaths

        # cut off the end of self
        sp = self.intersect(box2.path())[0]
        sp.sort()
        if sp:
            self.subpaths = self.split(sp[:1])[0].subpaths

    def shortenpath(self, dists):
        """shorten a path by the given distances"""

        # cut off the start of self
        # XXX should path.lentopar used here?
        center = (unit.topt(self.begin()[0]), unit.topt(self.begin()[1]))
        sp = self.intersect(path.circle_pt(center[0], center[1], dists[0]))[0]
        sp.sort()
        if sp:
            self.subpaths = self.split(sp[-1:])[1].subpaths

        # cut off the end of self
        center = (unit.topt(self.end()[0]), unit.topt(self.end()[1]))
        sp = self.intersect(path.circle_pt(center[0], center[1], dists[1]))[0]
        sp.sort()
        if sp:
            self.subpaths = self.split(sp[:1])[0].subpaths


################
## classes
################


class line_pt(connector_pt):

    def __init__(self, box1, box2, boxdists=[0,0]):

        self.box1 = box1
        self.box2 = box2

        connector_pt.__init__(self,
            [path.normsubpath([path.normline(*(self.box1.center+self.box2.center))], 0)])

        self.omitends(box1, box2)
        self.shortenpath(boxdists)


class arc_pt(connector_pt):

    def __init__(self, box1, box2, relangle=45,
                 absbulge=None, relbulge=None, boxdists=[0,0]):

        # the deviation of arc from the straight line can be specified:
        # 1. By an angle between a straight line and the arc
        #    This angle is measured at the centers of the box.
        # 2. By the largest normal distance between line and arc: absbulge
        #    or, equivalently, by the bulge relative to the length of the
        #    straight line from center to center.
        # Only one can be used.

        self.box1 = box1
        self.box2 = box2

        rel = (self.box2.center[0] - self.box1.center[0],
               self.box2.center[1] - self.box1.center[1])
        distance = hypot(*rel)

        # usage of bulge overrides the relangle parameter
        if relbulge is not None or absbulge is not None:
            relangle = None
            bulge = 0
            try: bulge += absbulge
            except: pass
            try: bulge += relbulge*distance
            except: pass

            try: radius = abs(0.5 * (bulge + 0.25 * distance**2 / bulge))
            except: radius = 10 * distance # default value for too straight arcs
            radius = min(radius, 10 * distance)
            center = 2.0*(radius-abs(bulge))/distance
            center *= 2*(bulge>0.0)-1
        # otherwise use relangle
        else:
            bulge=None
            try: radius = 0.5 * distance / abs(cos(0.5*math.pi - radians(relangle)))
            except: radius = 10 * distance
            try: center = tan(0.5*math.pi - radians(relangle))
            except: center = 0

        # up to here center is only the distance from the middle of the
        # straight connection
        center = (0.5 * (self.box1.center[0] + self.box2.center[0] - rel[1]*center),
                  0.5 * (self.box1.center[1] + self.box2.center[1] + rel[0]*center))
        angle1 = atan2(self.box1.center[1] - center[1], self.box1.center[0] - center[0])
        angle2 = atan2(self.box2.center[1] - center[1], self.box2.center[0] - center[0])

        # draw the arc in positive direction by default
        # negative direction if relangle<0 or bulge<0
        if (relangle is not None and relangle < 0) or (bulge is not None and bulge < 0):
            connector_pt.__init__(self,
                path.path(path.moveto_pt(*self.box1.center),
                          path.arcn_pt(center[0], center[1], radius, degrees(angle1), degrees(angle2))))
        else:
            connector_pt.__init__(self,
                path.path(path.moveto_pt(*self.box1.center),
                          path.arc_pt(center[0], center[1], radius, degrees(angle1), degrees(angle2))))

        self.omitends(box1, box2)
        self.shortenpath(boxdists)


class curve_pt(connector_pt):

    def __init__(self, box1, box2,
                 relangle1=45, relangle2=45,
                 absangle1=None, absangle2=None,
                 absbulge=0, relbulge=0.39, boxdists=[0,0]):

        # The deviation of the curve from a straight line can be specified:
        # A. By an angle at each center
        #    These angles are either absolute angles with origin at the positive x-axis
        #    or the relative angle with origin at the straight connection line
        # B. By the (expected) largest normal distance between line and arc: absbulge
        #    and/or by the (expected) bulge relative to the length of the
        #    straight line from center to center.
        # Here, we need both informations.
        #
        # a curve with relbulge=0.39 and relangle1,2=45 leads
        # approximately to the arc with angle=45

        self.box1 = box1
        self.box2 = box2

        rel = (self.box2.center[0] - self.box1.center[0],
               self.box2.center[1] - self.box1.center[1])
        distance = hypot(*rel)
        # absolute angle of the straight connection
        dangle = atan2(rel[1], rel[0])

        # calculate the armlength and absolute angles for the control points:
        # absolute and relative bulges are added
        bulge = abs(distance*relbulge + absbulge)

        if absangle1 is not None:
            angle1 = radians(absangle1)
        else:
            angle1 = dangle - radians(relangle1)
        if absangle2 is not None:
            angle2 = radians(absangle2)
        else:
            angle2 = dangle + radians(relangle2)

        # get the control points
        control1 = (cos(angle1), sin(angle1))
        control2 = (cos(angle2), sin(angle2))
        control1 = (self.box1.center[0] + control1[0] * bulge, self.box1.center[1] + control1[1] * bulge)
        control2 = (self.box2.center[0] - control2[0] * bulge, self.box2.center[1] - control2[1] * bulge)

        connector_pt.__init__(self,
               [path.normsubpath([path.normcurve(*(self.box1.center +
                                                   control1 +
                                                   control2 + self.box2.center))], 0)])

        self.omitends(box1, box2)
        self.shortenpath(boxdists)


class twolines_pt(connector_pt):

    def __init__(self, box1, box2,
                 absangle1=None, absangle2=None,
                 relangle1=None, relangle2=None, relangleM=None,
                 length1=None, length2=None,
                 bezierradius=None, beziersoftness=1,
                 arcradius=None,
                 boxdists=[0,0]):

        # The connection with two lines can be done in the following ways:
        # 1. an angle at each box-center
        # 2. two armlengths (if they are long enough)
        # 3. angle and armlength at the same box
        # 4. angle and armlength at different boxes
        # 5. one armlength and the angle between the arms
        #
        # Angles at the box-centers can be relative or absolute
        # The angle in the middle is always relative
        # lengths are always absolute

        self.box1 = box1
        self.box2 = box2

        begin = self.box1.center
        end = self.box2.center
        rel = (self.box2.center[0] - self.box1.center[0],
               self.box2.center[1] - self.box1.center[1])
        distance = hypot(*rel)
        dangle = atan2(rel[1], rel[0])

        # find out what arguments are given:
        if relangle1 is not None: relangle1 = radians(relangle1)
        if relangle2 is not None: relangle2 = radians(relangle2)
        if relangleM is not None: relangleM = radians(relangleM)
        # absangle has priority over relangle:
        if absangle1 is not None: relangle1 = dangle - radians(absangle1)
        if absangle2 is not None: relangle2 = math.pi - dangle + radians(absangle2)

        # check integrity of arguments
        no_angles, no_lengths=0,0
        for anangle in (relangle1, relangle2, relangleM):
            if anangle is not None: no_angles += 1
        for alength in (length1, length2):
            if alength is not None: no_lengths += 1

        if no_angles + no_lengths != 2:
            raise NotImplementedError, "Please specify exactly two angles or lengths"

        # calculate necessary angles and armlengths
        # always length1 and relangle1

        # the case with two given angles
        # use the "sine-theorem" for calculating length1
        if no_angles == 2:
            if relangle1 is None: relangle1 = math.pi - relangle2 - relangleM
            elif relangle2 is None: relangle2 = math.pi - relangle1 - relangleM
            elif relangleM is None: relangleM = math.pi - relangle1 - relangle2
            length1 = distance * abs(sin(relangle2)/sin(relangleM))
            middle = self._middle_a(begin, dangle, length1, relangle1)
        # the case with two given lengths
        # uses the "cosine-theorem" for calculating length1
        elif no_lengths == 2:
            relangle1 = acos((distance**2 + length1**2 - length2**2) / (2.0*distance*length1))
            middle = self._middle_a(begin, dangle, length1, relangle1)
        # the case with one length and one angle
        else:
            if relangle1 is not None:
                if length1 is not None:
                    middle = self._middle_a(begin, dangle, length1, relangle1)
                elif length2 is not None:
                    length1 = self._missinglength(length2, distance, relangle1)
                    middle = self._middle_a(begin, dangle, length1, relangle1)
            elif relangle2 is not None:
                if length1 is not None:
                    length2 = self._missinglength(length1, distance, relangle2)
                    middle = self._middle_b(end, dangle, length2, relangle2)
                elif length2 is not None:
                    middle = self._middle_b(end, dangle, length2, relangle2)
            elif relangleM is not None:
                if length1 is not None:
                    length2 = self._missinglength(distance, length1, relangleM)
                    relangle1 = acos((distance**2 + length1**2 - length2**2) / (2.0*distance*length1))
                    middle = self._middle_a(begin, dangle, length1, relangle1)
                elif length2 is not None:
                    length1 = self._missinglength(distance, length2, relangleM)
                    relangle1 = acos((distance**2 + length1**2 - length2**2) / (2.0*distance*length1))
                    middle = self._middle_a(begin, dangle, length1, relangle1)
            else:
                raise NotImplementedError, "I found a strange combination of arguments"

        connector_pt.__init__(self,
            path.path(path.moveto_pt(*self.box1.center),
                      path.lineto_pt(*middle),
                      path.lineto_pt(*self.box2.center)))

        self.omitends(box1, box2)
        self.shortenpath(boxdists)

    def _middle_a(self, begin, dangle, length1, angle1):
        a = dangle - angle1
        dir = cos(a), sin(a)
        return begin[0] + length1*dir[0], begin[1] + length1*dir[1]

    def _middle_b(self, end, dangle, length2, angle2):
        # a = -math.pi + dangle + angle2
        return self._middle_a(end, -math.pi+dangle, length2, -angle2)

    def _missinglength(self, lenA, lenB, angleA):
        # calculate lenC, where side A and angleA are opposite
        tmp1 = lenB * cos(angleA)
        tmp2 = sqrt(tmp1**2 - lenB**2 + lenA**2)
        if tmp1 > tmp2: return tmp1 - tmp2
        return tmp1 + tmp2



class line(line_pt):

    """a line is the straight connector between the centers of two boxes"""

    def __init__(self, box1, box2, boxdists=[0,0]):

        boxdists_pt = (_topt(helper.getitemno(boxdists, 0), default_type="v"),
                       _topt(helper.getitemno(boxdists, 1), default_type="v"))

        line_pt.__init__(self, box1, box2, boxdists=boxdists_pt)


class curve(curve_pt):

    """a curve is the curved connector between the centers of two boxes.
    The constructor needs both angle and bulge"""


    def __init__(self, box1, box2,
                 relangle1=45, relangle2=45,
                 absangle1=None, absangle2=None,
                 absbulge=0, relbulge=0.39,
                 boxdists=[0,0]):

        boxdists_pt = (_topt(helper.getitemno(boxdists, 0), default_type="v"),
                       _topt(helper.getitemno(boxdists, 1), default_type="v"))

        curve_pt.__init__(self, box1, box2,
                        relangle1=relangle1, relangle2=relangle2,
                        absangle1=absangle1, absangle2=absangle2,
                        absbulge=_topt(absbulge), relbulge=relbulge,
                        boxdists=boxdists_pt)

class arc(arc_pt):

    """an arc is a round connector between the centers of two boxes.
    The constructor gets
         either an angle in (-pi,pi)
         or a bulge parameter in (-distance, distance)
           (relbulge and absbulge are added)"""

    def __init__(self, box1, box2, relangle=45,
                 absbulge=None, relbulge=None, boxdists=[0,0]):

        boxdists_pt = (_topt(helper.getitemno(boxdists, 0), default_type="v"),
                       _topt(helper.getitemno(boxdists, 1), default_type="v"))

        arc_pt.__init__(self, box1, box2,
                      relangle=relangle,
                      absbulge=_topt(absbulge), relbulge=relbulge,
                      boxdists=boxdists_pt)


class twolines(twolines_pt):

    """a twolines is a connector consisting of two straight lines.
    The construcor takes a combination of angles and lengths:
      either two angles (relative or absolute)
      or two lenghts
      or one length and one angle"""

    def __init__(self, box1, box2,
                 absangle1=None, absangle2=None,
                 relangle1=None, relangle2=None, relangleM=None,
                 length1=None, length2=None,
                 bezierradius=None, beziersoftness=1,
                 arcradius=None,
                 boxdists=[0,0]):

        boxdists_pt = (_topt(helper.getitemno(boxdists, 0), default_type="v"),
                       _topt(helper.getitemno(boxdists, 1), default_type="v"))

        twolines_pt.__init__(self, box1, box2,
                       absangle1=absangle1, absangle2=absangle2,
                       relangle1=relangle1, relangle2=relangle2,
                       relangleM=relangleM,
                       length1=_topt(length1), length2=_topt(length2),
                       bezierradius=_topt(bezierradius), beziersoftness=1,
                       arcradius=_topt(arcradius),
                       boxdists=boxdists_pt)






#!/usr/bin/env python
# -*- coding: ISO-8859-1 -*-
#
#
# Copyright (C) 2002-2004 Jörg Lehmann <joergl@users.sourceforge.net>
# Copyright (C) 2002-2004 André Wobst <wobsta@users.sourceforge.net>
#
# This file is part of PyX (http://pyx.sourceforge.net/).
#
# PyX is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PyX is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyX; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


"""
(La)TeX interface of PyX

This module provides the classes tex and latex, which can be inserted into a
PyX canvas. The method (la)tex.text prints text, while (la)tex.textwd,
(la)tex.textht, and (la)tex.textdp appraise the width, height, and depth of a
text, respectively. The method (la)tex.define can be used to define macros in
(La)TeX.
"""

import os, string, tempfile, sys, md5, traceback, time, StringIO, re, atexit
import base, helper, unit, epsfile, color

sys.stderr.write("*** PyX Warning: the tex module is obsolete, consider the text module instead\n")

# Code snippets from former attrlist module (which has been removed from the
# CVS tree). We keep them here, until they are finally removed together with
# the tex module

class AttrlistError(base.PyXExcept):
    pass


class attrlist:
    def attrcheck(self, attrs, allowonce=(), allowmulti=()):
        hadonce = []
        for attr in attrs:
            for once in allowonce:
                if isinstance(attr, once):
                    if once in hadonce:
                        raise AttrlistError
                    else:
                        hadonce += [once]
                        break
            else:
                for multi in allowmulti:
                    if isinstance(attr, multi):
                        break
                else:
                    raise AttrlistError

    def attrgetall(self, attrs, get, default=helper.nodefault):
        first = 1
        for attr in attrs:
            if isinstance(attr, get):
                if first:
                    result = [attr]
                    first = 0
                else:
                    result.append(attr)
        if first:
            if default is helper.nodefault:
                raise AttrlistError
            else:
                return default
        return result

    def attrcount(self, attrs, check):
        return len(self.attrgetall(attrs, check, ()))

    def attrget(self, attrs, get, default=helper.nodefault):
        try:
            result = self.attrgetall(attrs, get)
        except AttrlistError:
            if default is helper.nodefault:
                raise AttrlistError
            else:
                return default
        if len(result) > 1:
            raise AttrlistError
        return result[0]

    def attrgetfirst(self, attrs, get, default=helper.nodefault):
        try:
            result = self.attrgetall(attrs, get)
        except AttrlistError:
            if default is helper.nodefault:
                raise AttrlistError
            else:
                return default
        return result[0]

    def attrgetlast(self, attrs, get, default=helper.nodefault):
        try:
            result = self.attrgetall(attrs, get)
        except AttrlistError:
            if default is helper.nodefault:
                raise AttrlistError
            else:
                return default
        return result[-1]

    def attrdel(self, attrs, remove):
        result = []
        for attr in attrs:
            if not isinstance(attr, remove):
                result.append(attr)
        return result


################################################################################
# TeX attributes
################################################################################

class _texattr:

    """base class for all TeX attributes"""

    pass


class fontsize(_texattr):
    
    """fontsize TeX attribute"""
    
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return self.value


fontsize.tiny = fontsize("tiny")
fontsize.scriptsize = fontsize("scriptsize")
fontsize.footnotesize = fontsize("footnotesize")
fontsize.small = fontsize("small")
fontsize.normalsize = fontsize("normalsize")
fontsize.large = fontsize("large")
fontsize.Large = fontsize("Large")
fontsize.LARGE = fontsize("LARGE")
fontsize.huge = fontsize("huge")
fontsize.Huge = fontsize("Huge")


class halign(_texattr):
    
    """tex horizontal align attribute"""

    def __init__(self, value):
        self.value = value

    def __cmp__(self, other):
        if other is None: return 1
        return cmp(self.value, other.value)

    __rcmp__ = __cmp__


halign.left   = halign("left")
halign.center = halign("center")
halign.right  = halign("right")

   
class valign(_texattr):

    """abstract tex vertical align attribute"""

    def __init__(self, hsize):
        self.hsize = hsize


class _valignvtop(valign):

    """tex top vertical align attribute"""

    pass


valign.top = _valignvtop


class _valignvbox(valign):

    """tex bottom vertical align attribute"""

    pass


valign.bottom = _valignvbox


class direction(_texattr):

    """tex output direction attribute"""

    def __init__(self, value):
        self.value = value

    def __str__(self):
        return "%.5f" % self.value



direction.horizontal = direction(0)
direction.vertical   = direction(90)
direction.upsidedown = direction(180)
direction.rvertical  = direction(270)


class style(_texattr):

    """tex style modification attribute"""

    def __init__(self, praefix, suffix):
        self.praefix = praefix
        self.suffix = suffix

    def ModifyCmd(self, cmd):
        return self.praefix + cmd + self.suffix


style.text = style("", "")
style.math = style("$\displaystyle{}", "$")


################################################################################
# TeX message handlers
################################################################################

class msghandler(_texattr):

    """abstract base class for tex message handlers
    
    A message handler has to provide a parsemsg method. It gets a string and
    returns a string. Within the parsemsg method the handler may remove any
    part of the message it is familiar with."""

    def removeemptylines(self, msg):
        """any message parser may use this method to remove empty lines"""

        msg = re.sub("^(\n)*", "", msg)
        msg = re.sub("(\n){3,}", "\n\n", msg)
        msg = re.sub("(\n)+$", "\n", msg)
        return msg


class _msghandlershowall(msghandler):

    """a message handler, which shows all messages"""

    def parsemsg(self, msg):
        return msg


msghandler.showall = _msghandlershowall()

class _msghandlerhideload(msghandler):

    """a message handler, which hides all messages inside proper '(filename' and ')'
    the string filename has to be a readable file"""

    def parsemsg(self, msg):
        depth = 0
        newstr = ""
        newlevel = 0
        for c in msg:
            if newlevel and (c in (list(string.whitespace) + ["(", ")"])):
                if filestr not in ("c", "C"):
                    if not len(filestr):
                        break
                    if not os.access(filestr,os.R_OK):
                        break
                newlevel = 0
            if c == "(":
                depth += 1
                filestr = ""
                newlevel = 1
            elif c == ")":
                depth -= 1
                if depth < 0:
                    break
            elif depth == 0:
                newstr += c
            else:
                filestr += c
        else:
            # replace msg only if loop was completed and no ")" is missing
            if depth == 0:
                msg = self.removeemptylines(newstr)
        return msg


msghandler.hideload = _msghandlerhideload()


class _msghandlerhidegraphicsload(msghandler):

    """a message handler, which hides all messages like '<filename>'
    the string filename has to be a readable file"""

    def parsemsg(self, msg):
        depth = 0
        newstr = ""
        for c in msg:
            if c == "<":
                depth += 1
                if depth > 1:
                    break
                filestr = ""
            elif c == ">":
                depth -= 1
                if depth < 0:
                    break
                if not os.access(filestr,os.R_OK):
                    newstr += "<" + filestr + ">"
            elif depth == 0:
                newstr += c
            else:
                filestr += c
        else:
            # replace msg only if loop was completed and no ">" missing
            if depth == 0:
                msg = self.removeemptylines(newstr)
        return msg


msghandler.hidegraphicsload = _msghandlerhidegraphicsload()


class _msghandlerhidefontwarning(msghandler):

    """a message handler, which hides LaTeX font warnings, e.g.
    Messages starting with 'LaTeX Font Warning: ' which might be
    continued on following lines by '(Font)              '"""

    def parsemsg(self, msg):
        msglines = string.split(msg, "\n")
        newmsglines = []
        fontwarning = 0
        for line in msglines:
            if fontwarning and line[:20] != "(Font)              ":
                fontwarning = 0
            if not fontwarning and line[:20] == "LaTeX Font Warning: ":
                fontwarning = 1
            if not fontwarning:
                newmsglines.append(line)
        newmsg = reduce(lambda x, y: x + y + "\n", newmsglines, "")
        return self.removeemptylines(newmsg)


msghandler.hidefontwarning = _msghandlerhidefontwarning()


class _msghandlerhidebuterror(msghandler):

    """a message handler, hides all messages whenever they do
    not contain a line starting with '! '"""

    def parsemsg(self, msg):
        # the "\n" + msg instead of msg itself is needed, if the message starts with "! "
        if string.find("\n" + msg, "\n! ") != -1:
            return msg
        else:
            return ""


msghandler.hidebuterror = _msghandlerhidebuterror()


class _msghandlerhideall(msghandler):

    """a message handler, which hides all messages"""

    def parsemsg(self, msg):
        return ""


msghandler.hideall = _msghandlerhideall()


################################################################################
# extent handlers
################################################################################

class missextents(_texattr):

    """abstract base class for handling missing extents

    A miss extent class has to provide a misshandler method."""


_missextentsreturnzero_report = 0
def _missextentsreturnzero_printreport():
    sys.stderr.write("""
pyx.tex: Some requested extents were missing and have been replaced by zero.
         Please run the file again to get correct extents.\n""")

class _missextentsreturnzero(missextents):

    def misshandler(self, texinstance):
        global _missextentsreturnzero_report
        if not _missextentsreturnzero_report:
            atexit.register(_missextentsreturnzero_printreport)
        _missextentsreturnzero_report = 1
        return map(lambda x: unit.t_pt(0), texinstance.BoxCmds[0].CmdExtents)


missextents.returnzero = _missextentsreturnzero()


class _missextentsreturnzeroquiet(missextents):

    def misshandler(self, texinstance):
        return map(lambda x: unit.t_pt(0), texinstance.BoxCmds[0].CmdExtents)


missextents.returnzeroquiet = _missextentsreturnzeroquiet()


class _missextentsraiseerror(missextents):

    def misshandler(self, texinstance):
        raise TexMissExtentError


missextents.raiseerror = _missextentsraiseerror()


class _missextentscreateextents(missextents):

    def misshandler(self, texinstance):
        if isinstance(texinstance, latex):
            storeauxfilename = texinstance.auxfilename
            texinstance.auxfilename = None
        texinstance.DoneRunTex = 0
        texinstance._run()
        texinstance.DoneRunTex = 0
        if isinstance(texinstance, latex):
            texinstance.auxfilename = storeauxfilename
        return texinstance.BoxCmds[0].Extents(texinstance.BoxCmds[0].CmdExtents,
                                              missextents.returnzero, texinstance)


missextents.createextents = _missextentscreateextents()


class _missextentscreateallextents(missextents):

    def misshandler(self, texinstance):
        if isinstance(texinstance, latex):
            storeauxfilename = texinstance.auxfilename
            texinstance.auxfilename = None
        texinstance.DoneRunTex = 0
        storeextents = texinstance.BoxCmds[0].CmdExtents[0]
        texinstance.BoxCmds[0].CmdExtents = [_extent.wd, _extent.ht, _extent.dp]
        texinstance._run()
        texinstance.BoxCmds[0].CmdExtents[0] = storeextents
        texinstance.DoneRunTex = 0
        if isinstance(texinstance, latex):
            texinstance.auxfilename = storeauxfilename
        return texinstance.BoxCmds[0].Extents(texinstance.BoxCmds[0].CmdExtents,
                                              missextents.returnzero, texinstance)


missextents.createallextents = _missextentscreateallextents()


################################################################################
# TeX exceptions
################################################################################

class TexExcept(base.PyXExcept):

    pass


class TexLeftParenthesisError(TexExcept):

    def __str__(self):
        return "no matching parenthesis for '{' found"


class TexRightParenthesisError(TexExcept):

    def __str__(self):
        return "no matching parenthesis for '}' found"


class TexHalignError(TexExcept):

    def __str__(self):
        return "unkown halign"


class TexValignError(TexExcept):

    def __str__(self):
        return "unkown valign"


class TexDefAfterBoxError(TexExcept):

    def __str__(self):
        return "definition commands not allowed after output commands"


class TexMissExtentError(TexExcept):

    def __str__(self):
        return "requested tex extent not available"


################################################################################
# modules internal stuff
################################################################################

class _extent:

    def __init__(self, value):
        self.value = value

    def __str__(self):
        return self.value


_extent.wd = _extent("wd")
_extent.ht = _extent("ht")
_extent.dp = _extent("dp")


class _TexCmd:

    """class for all user supplied commands"""

    PyxMarker = "PyxMarker"
    BeginPyxMarker = "Begin" + PyxMarker
    EndPyxMarker = "End" + PyxMarker

    def __init__(self, Marker, Stack, msghandlers):
        self.Marker = Marker
        self.Stack = Stack
        self.msghandlers = msghandlers

    def TexParenthesisCheck(self, Cmd):
        """check for proper usage of "{" and "}" in Cmd"""

        depth = 0
        esc = 0
        for c in Cmd:
            if c == "{" and not esc:
                depth = depth + 1
            if c == "}" and not esc:
                depth = depth - 1
                if depth < 0:
                    raise TexRightParenthesisError
            if c == "\\":
                esc = (esc + 1) % 2
            else:
                esc = 0
        if depth > 0:
            raise TexLeftParenthesisError

    def BeginMarkerStr(self):
        return "%s[%s]" % (self.BeginPyxMarker, self.Marker, )
    
    def WriteBeginMarker(self, file):
        file.write("\\immediate\\write16{%s}%%\n" % self.BeginMarkerStr())

    def EndMarkerStr(self):
        return "%s[%s]" % (self.EndPyxMarker, self.Marker, )

    def WriteEndMarker(self, file):
        file.write("\\immediate\\write16{%s}%%\n" % self.EndMarkerStr())

    def WriteError(self, msg):
        sys.stderr.write("Traceback (innermost last):\n")
        traceback.print_list(self.Stack)
        sys.stderr.write("(La)TeX Message:\n" + msg + "\n")

    def CheckMarkerError(self, file):
        """read markers and identify the message"""

        line = file.readline()
        while (line != "") and (line[:-1] != self.BeginMarkerStr()):
            line = file.readline()
        msg = ""
        line = file.readline()
        while (line != "") and (line[:-1] != self.EndMarkerStr()):
            msg = msg + line
            line = file.readline()
        if line == "":
            self.WriteError(msg)
            raise IOError
        else:
            # check if message can be ignored
            doprint = 0
            parsedmsg = msg
            for msghandler in self.msghandlers:
                parsedmsg = msghandler.parsemsg(parsedmsg)
            for c in parsedmsg:
                if c not in string.whitespace:
                    self.WriteError(parsedmsg)
                    break


class _DefCmd(_TexCmd):

    """definition commands"""

    def __init__(self, DefCmd, Marker, Stack, msghandlers):
        _TexCmd.__init__(self, Marker, Stack, msghandlers)
        self.TexParenthesisCheck(DefCmd)
        self.DefCmd = "%s%%\n" % DefCmd

    def write(self, file):
        self.WriteBeginMarker(file)
        file.write(self.DefCmd)
        self.WriteEndMarker(file)


class _CmdPut:

    """print parameters for a BoxCmd (data structure)"""

    def __init__(self, x, y, halign, direction, color):
        self.x = x
        self.y = y
        self.halign = halign
        self.direction = direction
        self.color = color


class _BoxCmd(_TexCmd):

    """BoxCmd (for printing text and getting extents)"""

    def __init__(self, DefCmdsStr, BoxCmd, style, fontsize, valign, Marker, Stack, msghandlers):
        _TexCmd.__init__(self, Marker, Stack, msghandlers)
        self.TexParenthesisCheck(BoxCmd)
        self.DefCmdsStr = DefCmdsStr
        self.BoxCmd = "{%s}%%\n" % BoxCmd # add another "{" to ensure, that everything goes into the Box
        self.CmdPuts = [] # list, where to put the command
        self.CmdExtents = [] # list, which extents are requested

        self.BoxCmd = style.ModifyCmd(self.BoxCmd)
        if valign is not None:
            if isinstance(valign, _valignvtop):
                self.BoxCmd = "\\linewidth%.5ftruept\\vtop{\\hsize\\linewidth{%s}}" % \
                               (unit.topt(valign.hsize) * 72.27/72, self.BoxCmd, )
            elif isinstance(valign, _valignvbox):
                self.BoxCmd = "\\linewidth%.5ftruept\\vbox{\\hsize\\linewidth{%s}}" % \
                               (unit.topt(valign.hsize) * 72.27/72, self.BoxCmd, )
            else:
                raise TexValignError
        self.BoxCmd = "\\setbox\\localbox=\\hbox{\\%s%s}%%\n" % (fontsize, self.BoxCmd, )

    def __cmp__(self, other):
        if other is None: return 1
        return cmp(self.BoxCmd, other.BoxCmd)

    __rcmp__ = __cmp__

    def write(self, file):
        self.WriteBeginMarker(file)
        file.write(self.BoxCmd)
        self.WriteEndMarker(file)
        for CmdExtent in self.CmdExtents:
            file.write("\\immediate\\write\\sizefile{%s:%s:%s:\\the\\%s\\localbox}%%\n" %
                       (self.MD5(), CmdExtent, time.time(), CmdExtent, ))
        for CmdPut in self.CmdPuts:

            file.write("{\\vbox to0pt{\\kern%.5ftruept\\hbox{\\kern%.5ftruept\\ht\\localbox0pt" %
                        (-CmdPut.y, CmdPut.x))

            if CmdPut.direction != direction.horizontal:
                file.write("\\special{ps: gsave currentpoint currentpoint translate " +
                           str(CmdPut.direction) + " neg rotate neg exch neg exch translate }")
            if CmdPut.color != color.gray.black:
                file.write("\\special{ps: ")
                CmdPut.color.outputPS(file)
                file.write(" }")
            if CmdPut.halign == halign.left:
                pass
            elif CmdPut.halign == halign.center:
                file.write("\kern-.5\wd\localbox")
            elif CmdPut.halign == halign.right:
                file.write("\kern-\wd\localbox")
            else:
                raise TexHalignError
            file.write("\\copy\\localbox")

            if CmdPut.color != color.gray.black:
                file.write("\\special{ps: ")
                color.gray.black.outputPS(file)
                file.write(" }")
            if CmdPut.direction != direction.horizontal:
                file.write("\\special{ps: currentpoint grestore moveto }")
            file.write("}\\vss}\\nointerlineskip}%\n")

    def MD5(self):
        """creates an MD5 hex string for texinit + Cmd"""

        h = string.hexdigits
        r = ''
        s = md5.md5(self.DefCmdsStr + self.BoxCmd).digest()
        for c in s:
            i = ord(c)
            r = r + h[(i >> 4) & 0xF] + h[i & 0xF]
        return r

    def Put(self, x, y, halign, direction, color):
        self.CmdPuts.append(_CmdPut(x, y, halign, direction, color))

    def Extents(self, extents, missextents, texinstance):
        """get sizes from previous LaTeX run"""

        for extent in extents:
            if extent not in self.CmdExtents:
                self.CmdExtents.append(extent)

        result = []
        for extent in extents:
            s = self.MD5() + ":" + str(extent)
            for size in texinstance.Sizes:
                if size[:len(s)] == s:
                    texpt = float(string.rstrip(size.split(":")[3][:-3]))
                    result.append(unit.t_pt(texpt * 72.0 / 72.27))
                    break
            else:
                break
        else:
            return result

        # extent was not found --- temporarily remove all other commands in
        # order to allow the misshandler to access everything it ever wants
        storeboxcmds = texinstance.BoxCmds
        storecmdputs = self.CmdPuts
        storecmdextents = self.CmdExtents
        texinstance.BoxCmds = [self, ]
        self.CmdPuts = []
        self.CmdExtents = extents
        try:
            result = missextents.misshandler(texinstance)
        finally:
            texinstance.BoxCmds = storeboxcmds
            self.CmdPuts = storecmdputs
            self.CmdExtents = storecmdextents
        return result


################################################################################
# tex, latex class
################################################################################

class _tex(base.PSCmd, attrlist):

    """major parts are of tex and latex class are shared and implemented here"""

    def __init__(self, defaultmsghandlers=msghandler.hideload,
                       defaultmissextents=missextents.returnzero,
                       texfilename=None):
        if isinstance(defaultmsghandlers, msghandler):
            self.defaultmsghandlers = (defaultmsghandlers,)
        else:
            self.defaultmsghandlers = defaultmsghandlers
        self.defaultmissextents = defaultmissextents
        self.texfilename = texfilename
        self.DefCmds = []
        self.DefCmdsStr = None
        self.BoxCmds = []
        self.DoneRunTex = 0

        if len(os.path.basename(sys.argv[0])):
            basename = os.path.basename(sys.argv[0])
            if basename[-3:] == ".py":
                basename = basename[:-3]
            self.SizeFileName = os.path.join(os.getcwd(), basename + ".size")
        else:
            self.SizeFileName = os.path.join(os.getcwd(), "pyxput.size")
        try:
            file = open(self.SizeFileName, "r")
            self.Sizes = file.readlines()
            file.close()
        except IOError:
            self.Sizes = [ ]

    def _execute(self, command):
        if os.system(command):
            sys.stderr.write("The exit code of the following command was non-zero:\n" + command +
"""\nUsually, additional information causing this trouble appears closeby.
However, you may check the origin by keeping all temporary files.
In order to achieve this, you have to specify a texfilename in the
constructor of the class pyx.(la)tex. You can then try to run the
command by yourself.\n""")

    def _createaddfiles(self, tempname):
        pass

    def _removeaddfiles(self, tempname):
        pass

    def _executetex(self, tempname):
        pass

    def _executedvips(self, tempname):
        self._execute("dvips -O0in,11in -E -o %(t)s.eps %(t)s.dvi > %(t)s.dvipsout 2> %(t)s.dvipserr" % {"t": tempname})

    def _run(self):
        """run (La)TeX & dvips, report errors, fill self.abbox & self.epsdata"""

        if self.DoneRunTex:
            return

        if self.texfilename:
            mktemp = str(self.texfilename)
        else:
            storetempdir = tempfile.tempdir
            tempfile.tempdir = os.curdir
            mktemp = tempfile.mktemp()
            tempfile.tempdir = storetempdir
        tempname = os.path.basename(mktemp)

        self._createaddfiles(tempname)

        texfile = open(tempname + ".tex", "w")

        texfile.write("\\nonstopmode%\n")
        texfile.write("\\def\PyX{P\\kern-.3em\\lower.5ex\\hbox{Y}\\kern-.18em X}%\n")
        texfile.write("\\newwrite\\sizefile%\n\\newbox\\localbox%\n\\newbox\\pagebox%\n")
        texfile.write("{\\catcode`\\~=12\\immediate\\openout\\sizefile=%s.size\\relax}%%\n" % tempname)

        for Cmd in self.DefCmds:
            Cmd.write(texfile)

        texfile.write("\\setbox\\pagebox=\\vbox{%\n")

        for Cmd in self.BoxCmds:
            Cmd.write(texfile)

        texfile.write("}\n\\immediate\\closeout\\sizefile\n\\shipout\\copy\\pagebox\n")
        texfile.write(self._endcmd())
        texfile.close()

        self._executetex(tempname)

        try:
            outfile = open(tempname + ".texout", "r")
            for Cmd in self.DefCmds + self.BoxCmds:
                Cmd.CheckMarkerError(outfile)
            outfile.close()
        except IOError:
            sys.stderr.write("""An unexpected error occured while reading the (La)TeX output.
May be, you just have no disk space available. Or something badly
in your commands caused (La)TeX to give up completely. Or your
(La)TeX installation might be broken at all.
You may try to check the origin by keeping all temporary files.
In order to achieve this, you have to specify a texfilename in the
constructor of the class pyx.tex. You can then try to run (La)TeX
by yourself.\n""")

        if not os.access(tempname + ".dvi", 0):
            sys.stderr.write("""Can't find the dvi file which should be produced by (La)TeX.
May be, you just have no disk space available. Or something badly
in your commands caused (La)TeX to give up completely. Or your
(La)TeX installation might be broken at all.
You may try to check the origin by keeping all temporary files.
In order to achieve this, you have to specify a texfilename in the
constructor of the class pyx.tex. You can then try to run (La)TeX
by yourself.\n""")

        else:
            self._executedvips(tempname)
            if not os.access(tempname + ".eps", 0):
                sys.stderr.write("""Error reading the eps file which should be produced by dvips.
May be, you just have no disk space available. Or something badly
in your commands caused dvips to give up completely. Or your
(La)TeX installation might be broken at all.
You may try to check the origin by keeping all temporary files.
In order to achieve this, you have to specify a texfilename in the
constructor of the class pyx.tex. You can then try to run dvips
by yourself.\n""")
            else:
                aepsfile = epsfile.epsfile(0, 0, tempname + ".eps", translatebbox=0, clip=0)
                self.abbox = aepsfile.bbox()
                self.aprolog = aepsfile.prolog()
                epsdatafile = StringIO.StringIO()
                aepsfile.outputPS(epsdatafile)
                self.epsdata = epsdatafile.getvalue()

        # merge new sizes

        OldSizes = self.Sizes

        try:
            NewSizeFile = open(tempname + ".size", "r")
            NewSizes = NewSizeFile.readlines()
            NewSizeFile.close()
        except IOError:
            NewSizes = []

        if (len(NewSizes) != 0) or (len(OldSizes) != 0):
            SizeFile = open(self.SizeFileName, "w")
            SizeFile.writelines(NewSizes)
            self.Sizes = NewSizes
            for OldSize in OldSizes:
                OldSizeSplit = OldSize.split(":")
                for NewSize in NewSizes:
                    if NewSize.split(":")[0:2] == OldSizeSplit[0:2]:
                        break
                else:
                    if time.time() < float(OldSizeSplit[2]) + 60*60*24:   # we keep size results for one day
                        SizeFile.write(OldSize)
                        self.Sizes.append(OldSize)

        if not self.texfilename:
            for suffix in ("tex", "log", "size", "dvi", "eps", "texout", "texerr", "dvipsout", "dvipserr", ):
                try:
                    os.unlink(tempname + "." + suffix)
                except:
                    pass

        self._removeaddfiles(tempname)
        self.DoneRunTex = 1

    def prolog(self):
        self._run()
        return self.aprolog

    def bbox(self):
        self._run()
        return self.abbox

    def outputPS(self, file):
        self._run()
        file.writelines(self.epsdata)

    def define(self, Cmd, *attrs):
        if len(self.BoxCmds):
            raise TexDefAfterBoxError
        self.DoneRunTex = 0
        self.attrcheck(attrs, (), (msghandler,))
        self.DefCmds.append(_DefCmd(Cmd,
                                    len(self.DefCmds)+ len(self.BoxCmds),
                                    traceback.extract_stack(),
                                    self.attrgetall(attrs, msghandler, self.defaultmsghandlers)))

    def _insertcmd(self, Cmd, *attrs):
        if not len(self.BoxCmds):
            self._beginboxcmds()
            self.DefCmdsStr = reduce(lambda x,y: x + y.DefCmd, self.DefCmds, "")
        mystyle = self.attrget(attrs, style, style.text)
        myfontsize = self.attrget(attrs, fontsize, fontsize.normalsize)
        myvalign = self.attrget(attrs, valign, None)
        mymsghandlers = self.attrgetall(attrs, msghandler, self.defaultmsghandlers)
        MyCmd = _BoxCmd(self.DefCmdsStr, Cmd, mystyle, myfontsize, myvalign,
                        len(self.DefCmds) + len(self.BoxCmds), traceback.extract_stack(), mymsghandlers)
        if MyCmd not in self.BoxCmds:
            self.BoxCmds.append(MyCmd)
        for Cmd in self.BoxCmds:
            if Cmd == MyCmd:
                UseCmd = Cmd    # we could use MyCmd directly if we have just inserted it before
                                # (that's due to the side effect, that append doesn't make a copy of the element,
                                # but we ignore this here -- we don't want to depend on this side effect)
        return UseCmd

    def _text(self, x, y, Cmd, *attrs):
        """print Cmd at (x, y) --- position parameters in postscipt points"""

        self.DoneRunTex = 0
        self.attrcheck(attrs, (style, fontsize, halign, valign, direction, color.color), (msghandler,))
        myhalign = self.attrget(attrs, halign, halign.left)
        mydirection = self.attrget(attrs, direction, direction.horizontal)
        mycolor = self.attrget(attrs, color.color, color.gray.black)
        self._insertcmd(Cmd, *attrs).Put(x * 72.27 / 72.0, y * 72.27 / 72.0, myhalign, mydirection, mycolor)

    def text(self, x, y, Cmd, *attrs):
        """print Cmd at (x, y)"""

        self._text(unit.topt(x), unit.topt(y), Cmd, *attrs)

    def textwd(self, Cmd, *attrs):
        """get width of Cmd"""

        self.DoneRunTex = 0
        self.attrcheck(attrs, (style, fontsize, missextents), (msghandler,))
        mymissextents = self.attrget(attrs, missextents, self.defaultmissextents)
        return self._insertcmd(Cmd, *attrs).Extents((_extent.wd, ), mymissextents, self)[0]

    def textht(self, Cmd, *attrs):
        """get height of Cmd"""

        self.DoneRunTex = 0
        self.attrcheck(attrs, (style, fontsize, valign, missextents), (msghandler,))
        mymissextents = self.attrget(attrs, missextents, self.defaultmissextents)
        return self._insertcmd(Cmd, *attrs).Extents((_extent.ht, ), mymissextents, self)[0]


    def textdp(self, Cmd, *attrs):
        """get depth of Cmd"""

        self.DoneRunTex = 0
        self.attrcheck(attrs, (style, fontsize, valign, missextents), (msghandler,))
        mymissextents = self.attrget(attrs, missextents, self.defaultmissextents)
        return self._insertcmd(Cmd, *attrs).Extents((_extent.dp, ), mymissextents, self)[0]


class tex(_tex):

    """tex class adds the specializations to _tex needed for tex"""

    def __init__(self, lfs="10pt", **addargs):
        _tex.__init__(self, **addargs)
        # XXX other ways for creating font sizes?
        try:
            LocalLfsName = str(lfs) + ".lfs"
            lfsdef = open(LocalLfsName, "r").read()
        except IOError:
            try:
                try:
                    SysLfsName = os.path.join(sys.prefix, "share", "pyx", str(lfs) + ".lfs")
                    lfsdef = open(SysLfsName, "r").read()
                except IOError:
                    SysLfsName = os.path.join(os.path.dirname(__file__), "lfs", str(lfs) + ".lfs")
                    lfsdef = open(SysLfsName, "r").read()
            except IOError:
                files = map(lambda x: x[:-4],
                            filter(lambda x: x[-4:] == ".lfs",
                                   os.listdir(".") +
                                   os.listdir(os.path.join(sys.prefix, "share", "pyx")),
                                   os.listdir(os.path.join(os.path.dirname(__file__), "lfs"))))
                raise IOError("file '%s.lfs' not found. Available latex font sizes:\n%s" % (lfs, files))
        self.define(lfsdef)
        self.define("\\newdimen\\linewidth%\n\\hsize0truein%\n\\vsize0truein%\n\\hoffset-1truein%\n\\voffset-1truein")

    def _beginboxcmds(self):
        pass

    def _endcmd(self):
        return "\\end\n"

    def _executetex(self, tempname):
        self._execute("tex %(t)s.tex > %(t)s.texout 2> %(t)s.texerr" % {"t": tempname})


class latex(_tex):

    """latex class adds the specializations to _tex needed for latex"""

    def __init__(self, docclass="article", docopt=None, auxfilename=None, **addargs):
        _tex.__init__(self, **addargs)
        self.auxfilename = auxfilename
        if docopt:
            self.define("\\documentclass[" + str(docopt) + "]{" + str(docclass) + "}")
        else:
            self.define("\\documentclass{" + str(docclass) + "}")
        self.define("\\hsize0truein%\n\\vsize0truein%\n\\hoffset-1truein%\n\\voffset-1truein")

    def _beginboxcmds(self):
        self.define("\\begin{document}")

    def _endcmd(self):
        return "\\end{document}\n"

    def _createaddfiles(self, tempname):
        if self.auxfilename is not None:
            writenew = 0
            try:
                os.rename(self.auxfilename + ".aux", tempname + ".aux")
            except OSError:
                writenew = 1
        else:
            writenew = 1
        if writenew:
            auxfile = open(tempname + ".aux", "w")
            auxfile.write("\\relax\n")
            auxfile.close()

    def _executetex(self, tempname):
        self._execute("latex %(t)s.tex > %(t)s.texout 2> %(t)s.texerr" % {"t": tempname})

    def _removeaddfiles(self, tempname):
        if self.auxfilename is not None:
            os.rename(tempname + ".aux", self.auxfilename + ".aux")
        else:
            os.unlink(tempname + ".aux")

